"""
Overview
--------

This script downloads sequence data for each gene of interest (ortholog) and
all the specified taxa. The files that are downloaded with this script can then
be used to create HMMER profiles, to search for similarity in a aminoacidic or
nucleotidic sequence.

Limitations
-----------

At the moment, the script uses Kegg Orthologs as the ortholog database.

.. warning::

    Some taxa may still black listed, because they are not relevant to the
    rumen microbiome. If you find such thing to occur to you, please contact me
    or open an issue on the repository.

Required Data
-------------

The script requires data from Kegg Orthologs and Uniprot to be downloaded,
before it can be used. The script **download_data** (:ref:`download-data`)
automates the process.

Workflow for Custom Profiles
----------------------------

.. blockdiag::

    {
        default_fontsize=16;
        default_textcolor = 'white';
        orientation = portrait;

        class mgkit [color = "#e41a1c", width=160, height=80];
        class data [color = "#4daf4a", width=160, height=80];
        class software [color = "#377eb8", width=160, height=80];
        class gff [color = "#984ea3", width=160, height=80];

        aa_seqs [label = "AA Sequences", class = data];
        aa_seqs_orig [label = "AA Sequences\\n(unknown)", class = data];
        nuc_seqs [label = "Nucleotide\\nSequences", class = data];
        alg_files [label = "Alignment\\nFiles", class = data];
        profiles [label = "Custom\\nProfiles", class = data];
        align [label = "Alignment\\n(clustalo)", class = software];
        Results [class = data];
        download_profiles, translate_seq [class = mgkit];
        download_profiles [fontsize=14];
        hmmbuild, hmmscan, nhmmscan [class = software];

        download_profiles -> aa_seqs -> align -> alg_files -> hmmbuild;
        hmmbuild -> profiles -> hmmscan -> Results;
        nuc_seqs -> translate_seq -> aa_seqs_orig -> hmmscan;
        profiles -> nhmmscan -> Results;
        nuc_seqs -> nhmmscan;
    }

The process of building the profiles to be used with HMMER is a step that
involves several tasks (illustrated in the Workflow above):

    #. download of data
    #. alignment of sequences
    #. conversion in HMMER profiles.

The first step involves, for all ortholog genes, to download all sequences
available for each taxon level of interest: this will produce a series of file
which contain the amino-acid sequences for each tuple gene-taxon. This sctipt,
`download_profiles` can be used. The aminoacidic sequences downloaded are then
aligned using Clustal Omega (or other) and for each alignment a profile is
built.

HMMER required the use of aminoacidic sequences, to be match against the
profiles. The `translate_seq` script can be used to translate nucleotidic
sequences into aminoacidic ones. However, the last version of HMMER should be
able to match nucleotidic sequences, but it was not tested by us. The example
Workflow above illustrate that.

Building profiles in this way, by going through all ortholog genes and choosing
the taxon level desired, opens the possibility of incrementally refining the
profiling of a metagenome without having to rerun all profiles again, as only
the new ones need to be run. Filtering the all the results is a much faster
operation.

Usage
-----

The default behaviour is to download all Kegg Orthologs for all taxa in the
given taxonomy. Taxa can be filtered by both lineage (e.g. archaea, carnivora,
etc.) and rank (e.g. genus, family, etc.). Another option is to specify the KO
and taxa IDs to download.

Taxa Filters
************

The way a taxon is specified is through a few different rules:

* specific **taxon ids** in uniprot
* a specific **taxon rank** (e.g.: genus, phylum, etc.)
* optional lineage filter: the lineage filter make sure that the name specified
  is included in the lineage attribute in the taxonomy.

As an example, if the rank chosen is genus, and the lineage option is
set to archaea, only the taxa whose rank is genus and that belong to the
archaea subtree will be downloaded::

    $ download_profiles -m EMAIL -r genus -l archaea mg_data/kegg.pickle \\
    -t mg_data/taxonomy.pickle

This allows to customise the level of specificity that we want in profiling and
make the process of downloading faster.
For metagenomic data, a good start is mixing different taxon ranks, using the
order or genus for the genes and then specifying a lineage of interest.

Because each profile is indipendent from each other, it's useful to start the
download with a certain rank and then run the profiling. During the profiling
a new download can be started and so on.

Specific Genes and Taxa
***********************

It is possible to download only specific taxa and KO and can be done using the
`-i` and `-ko` respectively. When `-ko` is used, loading Kegg Data with `-k` is
not required and it is up to the user to ensure the correct genes or taxa.

An example to download only KO from 3 different taxa::

    $ download_profiles -v -m EMAIL -ko K00016 -i 9611 9645 9682 \\
    -t mg_data/taxonomy.pickle

The same example using taxa filtering, instead (at the time of writing)::

    $ download_profiles -v -m EMAIL -ko K00016 -r genus -l carnivora \\
    -t mg_data/taxonomy.pickle

Changes
-------

.. versionchanged:: 0.2.1
    added `-ko` option, resolved issues caused by changes in library

"""

import os
import glob
import argparse
import urllib2
import logging
import itertools
import numpy
import pickle
import functools
from cStringIO import StringIO
import mgkit
from mgkit.io.fasta import load_fasta
from mgkit import logger
from mgkit import kegg
from mgkit.net import uniprot
from mgkit import taxon
import mgkit.simple_cache
from . import utils


is_ancestor = None

LOG = logging.getLogger(__name__)


def set_parser():
    "argument parser configuration"
    parser = argparse.ArgumentParser(
        description='Download KO sequences from Uniprot',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        default='profile_files',
        action='store',
        help='directory in which to store the downloaded files'
    )
    parser.add_argument(
        '-k',
        '--kegg-data',
        default='data/kegg.pickle',
        action='store',
        help='pickle file containing Kegg data'
    )
    parser.add_argument(
        '-m',
        '--email',
        required=True,
        action='store',
        help='email address to use for Uniprot communications'
    )
    parser.add_argument(
        '-t',
        '--taxon-data',
        default='data/taxonomy.pickle',
        action='store',
        help='pickle file containing taxonomy data'
    )
    parser.add_argument(
        '-r',
        '--taxon-rank',
        default=None,
        action='store',
        help='taxon rank to download'
    )
    parser.add_argument(
        '-l',
        '--lineage',
        default=None,
        action='store',
        help='lineage for filtering (e.g. archaea)'
    )
    parser.add_argument(
        '-i',
        '--taxon-id',
        action='store',
        nargs='+',
        type=int,
        help='id(s) of taxa to download. If specified take precedence over ' +
        'lineage-rank'
    )
    parser.add_argument(
        '-ko',
        '--ko-id',
        action='store',
        nargs='+',
        type=str,
        help='KO id(s)to download. If specified option -k is not needed'
    )
    parser.add_argument(
        '-R',
        '--only-reviewed',
        default=False,
        action='store_true',
        help="Only download reviewed sequences"
    )
    parser.add_argument(
        '-a',
        '--all-path',
        default=False,
        action='store_true',
        help='Download all KO from Kegg - exclude blacklist'
    )
    utils.add_basic_options(parser, manual=__doc__)

    return parser


def filter_taxonomy_by_lineage(taxonomy, taxon_ids, lineage):
    LOG.info("Filtering taxa by lineage: %s", lineage)
    lineage_id = taxonomy.find_by_name(lineage)

    if len(lineage_id) > 1:
        LOG.warning(
            'More than one taxon has the "%s" (%s) name, %d (%s) will be used',
            lineage,
            ", ".join(str(taxon_id) for taxon_id in lineage_id),
            lineage_id[0],
            taxonomy[lineage_id[0]].s_name
        )

    lineage_id = lineage_id[0]

    for taxon_id in taxon_ids:
        if is_ancestor(taxon_id, lineage_id):
            yield taxon_id


def filter_taxonomy_by_rank(taxonomy, taxon_ids, rank):
    LOG.info("Filtering taxa by rank: %s", rank)
    for taxon_id in taxon_ids:
        if taxonomy[taxon_id].rank == rank:
            yield taxon_id


def load_data(taxon_data, length_data_name):
    "Loads data for script"
    taxonomy = taxon.UniprotTaxonomy(taxon_data)

    try:
        LOG.info("Using profile's length data (%s)", length_data_name)
        length_data = pickle.load(open(length_data_name, 'r'))
    except IOError:
        LOG.info("Not found (%s), restarting", length_data_name)
        length_data = {}

    return taxonomy, length_data


def choose_taxa(taxonomy, options):
    "Returns the list of ids to look for in Uniprot"
    if options.taxon_id:
        taxon_ids = set(options.taxon_id)
        LOG.debug(taxon_ids)
    else:
        taxon_ids = set(x.taxon_id for x in taxonomy)
        LOG.info("Total number of taxa: %d", len(taxon_ids))
        if options.taxon_rank is not None:
            taxon_ids = set(
                filter_taxonomy_by_rank(
                    taxonomy, taxon_ids, options.taxon_rank.lower()
                )
            )
            LOG.info("Number reduced to %d", len(taxon_ids))
        if options.lineage is not None:
            taxon_ids = set(
                filter_taxonomy_by_lineage(
                    taxonomy, taxon_ids, options.lineage.lower()
                )
            )
    LOG.info("The number of taxa to download is %d", len(taxon_ids))
    return taxon_ids


def choose_ko_ids(kegg_data, options):
    "Returns the list of mapping ID->Name according to the options passed"
    if options.all_path:
        LOG.info("Dowloading all KOs, ignoring blacklist")
        ko_ids = kegg_data.get_ko_names()
    else:
        LOG.info("Dowloading only KOs in pathways not blacklisted")
        filt_ko = set(
            itertools.chain(
                *(kegg_data[path_id].genes.keys()
                  for path_id in kegg_data if path_id not in kegg.BLACK_LIST)
            )
        )
        ko_ids = dict(
            (ko_id, name)
            for ko_id, name in kegg_data.get_ko_names().items()
            if ko_id in filt_ko
        )
    return ko_ids


def map_ko_to_uniprot(ko_id, taxon_ids, reviewed, contact):
    "Returns the taxon IDs found in Uniprot for a specific id"
    query_base = "database:(type:ko {0}) AND ({1}){2}"

    query = query_base.format(
        ko_id,
        ' OR '.join(
            "taxonomy:{0}".format(taxon_id) for taxon_id in taxon_ids
        ),
        'AND reviewed:yes' if reviewed else ''
    )

    LOG.info("Downloading taxa list for KO: %s", ko_id)
    taxon_ids = uniprot.ko_to_mapping(
        ko_id, query,
        'organism-id',
        contact=contact
    )

    LOG.info("Found %d taxa for KO: %s", len(taxon_ids), ko_id)

    return taxon_ids


def filter_found_taxa(taxon_ids_found, taxon_ids, taxonomy):
    """
    Filter the taxa found in Uniprot, making sure that they at a lower level
    of those requested
    """
    taxon_ids_found = set(int(taxon_id) for taxon_id in taxon_ids_found)

    taxon_ids_download = set()

    for taxon_id_found in taxon_ids_found:
        if taxon_id_found not in taxonomy:
            LOG.warning(
                "Taxon id (%d) not found in local taxonomy, skipping",
                taxon_id_found
            )
            continue
        for taxon_id in taxon_ids:
            if is_ancestor(taxon_id_found, taxon_id):
                taxon_ids_download.add(taxon_id)
                break

    LOG.info('Downloading a total of %d taxa', len(taxon_ids_download))

    return taxon_ids_download


def download_ko_sequences(ko_id, taxon_ids, reviewed, contact):
    "Downloads the sequences associated to all taxon IDs provided"
    ko_seqs = {}

    for taxon_id in taxon_ids:
        key = (ko_id, taxon_id, reviewed)
        seqs = uniprot.get_sequences_by_ko(
            ko_id,
            taxon_id,
            contact,
            reviewed
        )
        if seqs.count('>') <= 1:
            LOG.warning(
                'Not enough Sequences (%d) for taxon %d',
                seqs.count('>'),
                taxon_id
            )
            continue

        LOG.debug(
            "Downloaded %d sequences for taxon id: %d",
            seqs.count('>'),
            taxon_id
        )
        ko_seqs[key] = seqs

    return ko_seqs


def write_ko_sequences(seqs, taxonomy, output_dir):
    "Writes fasta sequences to disc"
    for ko_id, taxon_id, reviewed in seqs:
        file_name = '{0}_{1}_{2}{3}.fa'.format(
            ko_id,
            taxon_id,
            taxonomy[taxon_id].s_name.replace(' ', '#'),
            '' if reviewed else '-nr'
        )
        file_handle = open(os.path.join(output_dir, file_name), 'w')
        file_handle.write(seqs[(ko_id, taxon_id, reviewed)])
        file_handle.close()


def add_profiles_to_length(seqs, length_data):
    "Adds the average profile length to the dictionary"
    for key in seqs:
        avg_len = numpy.fromiter(
            (len(seq[1]) for seq in load_fasta(StringIO(seqs[key]))),
            dtype=int
        ).mean()
        length_data[key] = avg_len


def main():
    "Main function"
    options = set_parser().parse_args()

    logger.config_log(options.verbose)
    ret_value = 0

    contact = options.email
    reviewed = options.only_reviewed

    output_dir = options.output_dir
    try:
        LOG.info("Writing fasta files to %s", output_dir)
        os.mkdir(output_dir)
    except OSError:
        LOG.info("Output directory %s already present", output_dir)

    length_data_name = options.output_dir.rstrip('/') + '-length.pickle'

    taxonomy, length_data = load_data(
        options.taxon_data,
        length_data_name
    )

    global is_ancestor

    is_ancestor = mgkit.simple_cache.memoize(
        functools.partial(
            taxon.is_ancestor,
            taxonomy
        )
    )

    taxon_ids = choose_taxa(taxonomy, options)
    if len(taxon_ids) == 0:
        utils.exit_script("No taxa to download (0)", 3)

    if options.ko_id:
        ko_ids = set(ko_id.upper() for ko_id in options.ko_id)
    else:
        kegg_data = kegg.KeggData(options.kegg_data)
        ko_ids = choose_ko_ids(kegg_data, options)

    for idx, ko_id in enumerate(sorted(ko_ids)):
        if ko_id in (key[0] for key in length_data):
            LOG.info(
                "(%05d/%05d) Skipped %s, already downloaded",
                idx + 1,
                len(ko_ids),
                ko_id
            )
            continue
        LOG.info("(%05d/%05d) Downloading %s", idx + 1, len(ko_ids), ko_id)
        try:
            taxon_ids_found = map_ko_to_uniprot(ko_id, taxon_ids, reviewed,
                                                contact)
            taxon_ids_found = filter_found_taxa(
                taxon_ids_found,
                taxon_ids,
                taxonomy
            )
            seqs = download_ko_sequences(
                ko_id,
                taxon_ids_found,
                reviewed,
                contact
            )
        except urllib2.HTTPError, error:
            LOG.error(
                "Couldn't download sequences for %s: %s",
                ko_id,
                str(error)
            )
            ret_value = 1
            continue

        try:
            write_ko_sequences(seqs, taxonomy, output_dir)
            add_profiles_to_length(seqs, length_data)
        except IOError, error:
            LOG.error("Couldn't write sequences for %s", ko_id)
            LOG.error(str(error))
            file_list = glob.glob(
                os.path.join(
                    output_dir,
                    '{0}*'.format(ko_id)
                )
            )
            for fname in file_list:
                LOG.error("-> %s", fname)
                os.remove(fname)
            ret_value = 2
            continue
        else:
            LOG.debug("Profiles for %s saved", ko_id)
            pickle.dump(length_data, open(length_data_name, 'w'))

    return ret_value

if __name__ == '__main__':
    main()



#!python
"""

passare in summarise_gff con relative tabelle?
    coverage_of_taxa-min10
    number_of_taxa_per_gene
    number_of_genes_per_taxon

OK:
    fare qui
    taxa-all-sorted_by_median-min10
    boxplot-all_genes-sorted_by_median-min10

    wilcoxon test:
        geni
        taxa
--------------------------------------------

wilcoxon pairwise test:
    boxplot per taxon con tutti i geni significativi per quel taxon (se e' uno
    solo il gene evitare tight layout)
    barchart delle categorie per taxon

clustering:
    usare PCA per decidere le componenti in automatico? riservare per
    successivo sviluppo o lasciare primo quartile e media come misure?
    - non fare per il momento. Se fatto, meglio usare le tre misure (primo
    quartile, media e secondo quartile) ed un plot 3D

    i dati dnds derivano da tutti i geni hanno almeno 10 repliche che superano
    il test wilcoxon pairwise (fatto)
    poter scegliere se usare solo quelli significativi dal test wilcoxon o meno
    workflow dN/dS: saltare profile con solo una sequenza - fare successivamente

    costruire script per generare in automatico i valori di dN/dS con crann
        - generare il plot
        - generare la lista dei geni-taxa e loro appartenenza al cluster in un
          foglio excel

term enrichment:
    integrare goatools?
    ouput dei dati comunque nel formato richiesto
    controllare invece possibili altri tool usati da altre persone

"""

import cPickle
import argparse
import logging
import os
import numpy
import mgkit
from mgkit import taxon
from mgkit import logger
from mgkit.taxon import BLACK_LIST
from mgkit.workflow.snps_wf import load_sample_data, write_single_values
from mgkit.workflow.snps_wf import write_gt_values, plot_results_categories
from mgkit.workflow.snps_wf import get_sample_ratios, plot_results_boxplot
from mgkit.snps import combine_genes_by_feature
from mgkit.snps import convert_dict_for_pairwise_wilcoxon
from mgkit.snps import wilcoxon_test_ratio_pairwise
from mgkit.snps import convert_dict_for_wilcoxon, wilcoxon_test_ratio
from mgkit.plots import get_taxon_colors
from mgkit.utils.dictionary import split_dictionary_by_value

#if r_utils modules cannot be imported gives a warning
try:
    from mgkit.utils.r_func import correct_pvalues
except ImportError:
    correct_pvalues = None


LOG = logging.getLogger(__name__)


def make_dir(dir_name):
    try:
        LOG.info("Writing ouput files to %s", dir_name)
        os.mkdir(dir_name)
    except OSError:
        LOG.warning("Output directory %s already present", dir_name)


def get_taxa_colors(taxa, taxon_data):
    root_map = taxon.load_taxonomy_map(taxon_data)
    color_map = get_taxon_colors(taxa, root_map)
    return color_map


def load_mapping_data(map_files):
    pass


def gene_taxon_clustering(gt_values):
    pass


def taxon_analysis(options):
    LOG.info("Starting Taxon analysis")
    sample_data = load_sample_data(options.sample_data)
    make_dir(options.out_dir)
    t_values = combine_genes_by_feature(
        sample_data,
        'taxon',
        options.min_cov
    )
    s_values = get_sample_ratios(sample_data, options.min_cov)
    t_values = convert_dict_for_wilcoxon(t_values)
    #distribuzione campio
    sign_taxa = wilcoxon_test_ratio(
        s_values,
        t_values,
        corr_func=correct_pvalues if options.no_corr else None,
        corr_method=options.corr_method,
        min_num=options.min_num,
        threshold=options.threshold
    )
    LOG.info("Number of significant taxa: %d", len(sign_taxa))
    write_single_values(
        t_values,
        sign_taxa,
        'taxon',
        os.path.join(
            options.out_dir,
            'taxon_values.tab'
        )
    )
    color_map = get_taxa_colors(t_values, options.taxon_data)
    plot_base = os.path.join(options.out_dir, 'taxa_plot-')
    plot_results_boxplot('taxa', s_values, t_values, sign_taxa,
                         options.label_size, options.fig_ratio,
                         sign_taxa.index, plot_base, colors=color_map)


def gene_analysis(options):
    LOG.info("Starting Gene analysis")
    sample_data = load_sample_data(options.sample_data)
    make_dir(options.out_dir)
    g_values = combine_genes_by_feature(
        sample_data,
        'gene',
        options.min_cov
    )
    g_values = convert_dict_for_wilcoxon(g_values)
    s_values = get_sample_ratios(sample_data, options.min_cov)
    #distribuzione campio
    sign_genes = wilcoxon_test_ratio(
        s_values,
        g_values,
        corr_func=correct_pvalues if options.no_corr else None,
        corr_method=options.corr_method,
        min_num=options.min_num,
        threshold=options.threshold
    )
    LOG.info("Number of significant genes: %d", len(sign_genes))
    write_single_values(
        g_values,
        sign_genes,
        'gene',
        os.path.join(
            options.out_dir,
            'gene_values.tab'
        )
    )
    plot_base = os.path.join(options.out_dir, 'genes_plot-')
    plot_results_boxplot('genes', s_values, g_values, sign_genes,
                         options.label_size, options.fig_ratio,
                         sign_genes.index, plot_base)
    sign_genes_low, sign_genes_high = split_dictionary_by_value(
        g_values,
        numpy.median(s_values),
        key_filter=sign_genes.index
    )
    plot_base = os.path.join(options.out_dir, 'genes_categories-')
    plot_results_categories(g_values,
                            sign_genes_low,
                            sign_genes_high,

                            options.label_size,
                            aspect=options.fig_ratio,
                            colors=('g', 'r', 'b'),
                            base_name=plot_base
                            )


def gene_taxon_analysis(options):
    LOG.info("Starting Gene-Taxon analysis")
    sample_data = load_sample_data(options.sample_data)
    make_dir(options.out_dir)
    gt_values = combine_genes_by_feature(
        sample_data,
        'gene-taxon',
        options.min_cov
    )
    gt_values = convert_dict_for_pairwise_wilcoxon(gt_values)
    sign_genes = wilcoxon_test_ratio_pairwise(
        gt_values,
        corr_func=correct_pvalues if options.no_corr else None,
        corr_method=options.corr_method,
        min_num=options.min_num,
        threshold=options.threshold
    )
    LOG.info("Number of significant genes: %d", len(sign_genes))
    dnds = None
    if options.dnds_file is not None:
        LOG.info("Loading dN/dS data from file %s", options.dnds_file.name)
        dnds = cPickle.load(options.dnds_file)
        # gene_taxon_clustering(gt_values)

    write_gt_values(
        gt_values, sign_genes, os.path.join(
            options.out_dir,
            'gene-taxon_values.tab'
        ),
        dnds=dnds
    )


def add_stats_options(parser):
    parser.add_argument(
        '-c',
        '--no-corr',
        default=True,
        action='store_false',
        help='Disables p-value correction using R'
    )
    parser.add_argument(
        '-t',
        '--corr-method',
        default='bonferroni',
        action='store',
        help='Method used for p-value correction: bonferroni, BH, etc.'
    )
    parser.add_argument(
        '--min-num',
        default=10,
        action='store',
        type=int,
        help='minimum number of replicates accepted for a gene/taxon'
    )
    parser.add_argument(
        '--threshold',
        default=0.05,
        action='store',
        type=int,
        help='p-value threshold for acceptance'
    )


def add_graph_options(parser):
    """
    Sets command line arguments parser
    """
    parser.add_argument(
        '--fig-ratio',
        default=0.3,
        action='store',
        type=float,
        help='Figure aspect ratio'
    )
    parser.add_argument(
        '--label-size',
        default=6,
        action='store',
        type=int,
        help='X axis ticks font size. 0 means no labels'
    )


def add_general_options(parser):
    """
    Sets command line arguments parser
    """
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_const',
        const=logging.DEBUG,
        default=logging.INFO,
        help='more verbose'
    )
    parser.add_argument(
        '-m',
        '--min-cov',
        default=4,
        action='store',
        type=int,
        help='Minimum per sample coverage accepted'
    )
    parser.add_argument(
        '-b',
        '--black-list',
        nargs='*',
        default=BLACK_LIST,
        # action='store',
        help='Taxa to be excluded from analysis'
    )


def add_taxon_command(subparser):
    parser = subparser.add_parser(
        'taxon',
        help='Taxon analysis',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '--taxon-data',
        default='data/taxonomy.pickle',
        action='store',
        help='Taxonomy data, used to color plots'
    )
    parser.set_defaults(func=taxon_analysis)
    add_general_options(parser)
    add_stats_options(parser)
    add_command_commons(parser)


def add_gene_command(subparser):
    parser = subparser.add_parser(
        'gene',
        help='Gene analysis',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '--eggnog-data',
        default=None,
        action='store',
        help='eggNOG data, used for mappings'
    )
    parser.add_argument(
        '--cazy-data',
        default=None,
        action='store',
        help='CaZy data, used for mappings'
    )
    parser.add_argument(
        '-g',
        '--go-data',
        default=None,
        action='store',
        help='GO data, used for mappings'
    )
    parser.set_defaults(func=gene_analysis)
    add_general_options(parser)
    add_stats_options(parser)
    add_command_commons(parser)


def add_gene_taxon_command(subparser):
    parser = subparser.add_parser(
        'gene-taxon',
        help='Gene-Taxon analysis',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.set_defaults(func=gene_taxon_analysis)
    parser.add_argument(
        '-d',
        '--dnds-file',
        type=argparse.FileType('r'),
        action='store',
        default=None,
        help='File containing data for dN/dS values'
    )
    add_general_options(parser)
    add_stats_options(parser)
    add_command_commons(parser)


def add_command_commons(parser):
    parser.add_argument(
        'sample_data',
        type=argparse.FileType('r'),
        action='store',
        help='Saved data from VCF and SNPDat parsing'
    )
    parser.add_argument(
        'out_dir',
        action='store',
        help='Directory to which results are saved'
    )
    add_graph_options(parser)


def main():
    parser = argparse.ArgumentParser(
        description='SNPs analysis',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {0}'.format(mgkit.__VERSION__)
    )

    subparser = parser.add_subparsers(help='Possible analysis')
    add_gene_taxon_command(subparser)
    add_taxon_command(subparser)
    add_gene_command(subparser)

    options = parser.parse_args()

    logger.config_log(options.verbose)

    if correct_pvalues is None:
        LOG.warning(
            "Error in importing r_utils, p-value correction won't be performed"
        )

    options.func(options)

if __name__ == '__main__':
    main()



"""
.. versionadded:: 0.3.0

Scripts that includes some functionality to help use FASTA files with the
framework

split command
*************

Used to split a fasta file into smaller fragments

Changes
*******

.. versionadded:: 0.3.0

"""

from __future__ import division
import argparse
import logging

import mgkit
from . import utils
from mgkit.io import fasta

LOG = logging.getLogger(__name__)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='FASTA utilities',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    utils.add_basic_options(parser, manual=__doc__)

    subparsers = parser.add_subparsers()

    parser_split = subparsers.add_parser(
        'split',
        help='Splits a FASTA file in a number of fragments'
    )

    set_split_parser(parser_split)
    utils.add_basic_options(parser_split, manual=__doc__)

    return parser


def set_split_parser(parser):
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input FASTA file, defaults to stdin'
    )
    parser.add_argument(
        '-p',
        '--prefix',
        type=str,
        default='split',
        help='Prefix for the file name in output'
    )
    parser.add_argument(
        '-n',
        '--number',
        type=int,
        default=10,
        help='Number of chunks into which split the FASTA file'
    )
    parser.add_argument(
        '-z',
        '--gzip',
        action='store_true',
        default=False,
        help='gzip output files'
    )
    parser.set_defaults(func=split_command)


def split_command(options):
    LOG.info(
        "Splitting FASTA into %d chunks with prefix %s",
        options.number,
        options.prefix
    )

    name_mask = "%s-{0:05}.fa" % options.prefix
    if options.gzip:
        name_mask += '.gz'
        LOG.info("Output files will be compressed (gzip)")

    fasta.split_fasta_file(
        options.input_file,
        name_mask,
        num_files=options.number
    )


def main():
    "Main function"

    options = set_parser().parse_args()

    mgkit.logger.config_log(options.verbose)
    options.func(options)



"""
.. versionadded:: 0.1.14

EC mappings
"""

import re
from ..io import open_file

LEVEL1_NAMES = {
    1: 'oxidoreductases',
    2: 'transferases',
    3: 'hydrolases',
    4: 'lyases',
    5: 'isomerases',
    6: 'ligases'
}
"""
Top level classification names
"""

ENZCLASS_REGEX = r"^(\d)\. ?([\d-]+)\. ?([\d-]+)\. ?([\d-]+) +(.+)\."
"""
Used to get the description for the higher level enzyme classes from the file
*enzclass.txt* on `expasy <http://expasy.org>`_
"""


def parse_expasy_file(file_name):
    """
    Used to load enzyme descriptions from the file *enzclass.txt* on
    `expasy <http://expasy.org>`_.

    The FTP url for enzclass.txt is:
    `<ftp://ftp.expasy.org/databases/enzyme/enzclass.txt>`_
    """
    labels = {}

    for line in open_file(file_name, mode='r'):
        match = re.search(ENZCLASS_REGEX, line)

        if match is None:
            continue

        ec = '.'.join(value for value in match.groups()[:4] if value != '-')

        labels[ec] = match.group(5)

    return labels


def get_enzyme_level(ec, level=4):
    """
    .. versionadded:: 0.1.14

    Returns an enzyme class at a specific level , between 1 and 4 (by default
    the most specific, 4)

    Arguments:
        ec (str): a string representing an EC number (e.g. 1.2.4.10)
        level (int): from 1 to 4, to get a different level specificity of in
            the enzyme classification

    Returns:
        str: the EC number at the requested specificity

    Example:
        >>> from mgkit.mappings.enzyme import get_enzyme_level
        >>> get_enzyme_level('1.1.3.4', 1)
        '1'
        >>> get_enzyme_level('1.1.3.4', 2)
        '1.1'
        >>> get_enzyme_level('1.1.3.4', 3)
        '1.1.3'
        >>> get_enzyme_level('1.1.3.4', 4)
        '1.1.3.4'
    """
    return '.'.join(ec.split('.')[:level])


def change_mapping_level(ec_map, level=3):
    """
    .. versionadded:: 0.1.14

    Given a dictionary, whose values are dictionaries, in which a key is named
    *ec* and its value is an iterable of EC numbers, returns an iterator that
    can be used to build a dictionary with the same top level keys and the
    values are sets of the transformed EC numbers.

    Arguments:
        ec_map (dict): dictionary generated by
            :func:`mgkit.net.uniprot.get_gene_info`
        level (int): number from 1 to 4, to specify the level of the mapping,
            passed to :func:`get_enzyme_level`

    Yields:
        tuple: a tuple (gene_id, set(ECs)), which can be passed to *dict* to
        make a dictionary

    Example:
        >>> from mgkit.net.uniprot import get_gene_info
        >>> from mgkit.mappings.enzyme import change_mapping_level
        >>> ec_map = get_gene_info('Q9HFQ1', columns='ec')
        {'Q9HFQ1': {'ec': '1.1.3.4'}}
        >>> dict(change_mapping_level(ec_map, level=2))
        {'Q9HFQ1': {'1.1'}}

    """
    for gene_id, ecdict in ec_map.iteritems():
        try:
            ec_list = ecdict['ec']
        except KeyError:
            continue

        if isinstance(ec_list, str):
            ec_list = [ec_list]

        yield gene_id, set(
            get_enzyme_level(ec_id, level=level) for ec_id in ec_list
        )


def get_mapping_level(ec_map, level=3):
    """
    .. versionadded:: 0.3.0

    Given a dictionary, whose values are iterable of EC numbers, returns an
    iterator that can be used to build a dictionary with the same top level
    keys and the values are sets of the transformed EC numbers.

    Arguments:
        ec_map (dict): dictionary genes to EC
        level (int): number from 1 to 4, to specify the level of the mapping,
            passed to :func:`get_enzyme_level`

    Yields:
        tuple: a tuple (gene_id, set(ECs)), which can be passed to *dict* to
        make a dictionary
    """
    for gene_id, ec_list in ec_map.iteritems():

        if not ec_list:
            continue

        if isinstance(ec_list, str):
            ec_list = [ec_list]

        yield gene_id, set(
            get_enzyme_level(ec_id, level=level) for ec_id in ec_list
        )


def get_enzyme_full_name(ec_id, ec_names, sep=', '):
    """
    .. versionadded:: 0.2.1

    From a EC identifiers and a dictionary of names builds a comma separated
    name (by default) that identifies the function of the enzyme.

    Arguments:
        ec_id (str): EC identifier
        ec_names (dict): a dictionary of names that can be produced using
            :func:`parse_expasy_file`
        sep (str): string used to join the names

    Returns:
        str: the enzyme classification name
    """

    name_list = []

    while True:
        try:
            name_list.append(
                ec_names[ec_id]
            )
        except KeyError:
            pass

        ec_id = '.'.join(ec_id.split('.')[:-1])

        if not ec_id:
            break

    return sep.join(reversed(name_list))



"""
The scripts downloads the data that is used by the framework for some of its
functions. It's mostly a shortcut to call the download_data function that is
present in every module that is in the package mappings and in the kegg module.

The only option required, is the email contact for the person using the script;
this is used to make sure that the API requirements in Uniprot are fullfilled
and they can contact the person using the script is any problem arise.

.. note::

    The default behavior is to download first the taxonomy data form Uniprot,
    Kegg and additional mapping data.

Taxonomy
********

It is downloaded from Uniprot and build a data structure that is used by
several scripts and function in the package. The download can take some time.

.. warning::

    if only the taxonomy is to be downloaded, both the `-x` and `-p` options
    must be passed to the script.

Kegg Onthologs
**************

The Kegg data is the only "required" data at the moment, because it's used to
download the sequence data (via the donwload_profiles script) for the
profiling. It is the only data that can't be saved unless it's fully
downloaded.

Kegg data is required by the mappers currently supported, and its download
takes longer. The mappers handle timeouts and if exceptions are raised the data
is saved and the download is resumed when the script is started again.

Other Mappings
**************

The other mappings (from KO) are downloaded by default and this can be excluded
by using the `-p` option. Mappings for Gene Onthology, eggNOG and CaZy are
downloaded.

As the download of the mappings can take a lot of time, or break because of the
number of requests to the web sites, checkpoints are saved often, so it can
resumed at a later time.

.. warning::

    the Gene Onthology module has specific requirements, so if they are not
    downloaded

"""
import os
import argparse
import logging
import mgkit
import mgkit.logger
import mgkit.kegg
import mgkit.mappings.cazy
import mgkit.mappings.eggnog
import mgkit.taxon
import mgkit.net
from . import utils

try:
    import mgkit.mappings.go
    SKIP_GO = False
except ImportError:
    SKIP_GO = True

LOG = logging.getLogger(__name__)

TAXONONY_URL = "http://www.uniprot.org/taxonomy/?query=*&format=tab&" + \
    "compress=no"


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='SNPs analysis, requires a vcf file and SNPDat results',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        default='mg_data',
        action='store',
        help='Ouput directory'
    )
    parser.add_argument(
        '-k',
        '--kegg',
        default='kegg.pickle',
        action='store',
        help='Kegg data file name'
    )
    parser.add_argument(
        '-c',
        '--cazy',
        default='cazy.pickle',
        action='store',
        help='CaZy data file name'
    )
    parser.add_argument(
        '-g',
        '--go',
        default='go.pickle',
        action='store',
        help='Gene Onthology data file name'
    )
    parser.add_argument(
        '-e',
        '--eggnog',
        default='eggnog.pickle',
        action='store',
        help='eggNOG data file name'
    )
    parser.add_argument(
        '-t',
        '--taxonomy',
        default='taxonomy.pickle',
        action='store',
        help='Taxonomy data file name'
    )
    parser.add_argument(
        '-p',
        '--no-mappings',
        default=True,
        action='store_false',
        help='Use to not download Mapping data'
    )
    parser.add_argument(
        '-x',
        '--only-taxonomy',
        default=True,
        action='store_false',
        help='Use to only download Taxonomy data, no Kegg data'
    )
    parser.add_argument(
        '-m',
        '--email',
        required=True,
        action='store',
        help='email address to use for Uniprot communications'
    )
    utils.add_basic_options(parser)

    return parser


def main():
    "Main function"
    options = set_parser().parse_args()

    # configs log and set log level
    mgkit.logger.config_log(options.verbose)

    try:
        os.mkdir(options.output_dir)
    except OSError:
        LOG.info('Directory exists')

    LOG.info("Downloading Uniprot Taxonomy data")
    taxonomy_path = os.path.join(options.output_dir, options.taxonomy)
    if not os.path.exists(taxonomy_path):
        taxonomy = mgkit.taxon.UniprotTaxonomy()
        taxonomy.read_taxonomy(mgkit.net.url_open(TAXONONY_URL))
        taxonomy.save_data(taxonomy_path)

    if options.only_taxonomy:
        LOG.info("Downloading Kegg data")
        kegg_path = os.path.join(options.output_dir, options.kegg)
        if not os.path.exists(kegg_path):
            mgkit.kegg.download_data(fname=kegg_path, contact=options.email)

    if options.no_mappings:
        LOG.info("Downloading additional mappings")
        mgkit.mappings.cazy.download_data(
            options.email,
            kegg_data=kegg_path,
            cazy_data=os.path.join(options.output_dir, options.cazy)
        )
        mgkit.mappings.eggnog.download_data(
            options.email,
            kegg_data=kegg_path,
            eggnog_data=os.path.join(options.output_dir, options.eggnog)
        )
        if SKIP_GO:
            LOG.warning(
                "Cannot import mgkit.mappings.go, will be skipped"
            )
            return 1
        mgkit.mappings.go.download_data(
            options.email,
            kegg_data=kegg_path,
            go_data=os.path.join(options.output_dir, options.go)
        )

    LOG.info("Download completed")

    return 0

if __name__ == '__main__':
    main()



"""
Extract information from GFF files

sequence command
****************

Used to extract the nucleotidic sequences from GFF annotations. It requires the
*fasta* file containing the sequences referenced in the GFF `seq_id` attribute
(first column of the raw GFF).

The sequnces extract have as identifier the `uid` stored in the GFF file and by
default the sequnece is not reverse complemented if the annotation is on the
**-** strand, but this can be changed by using the `-r` option.

The sequences are wrapped at 60 characters, as per FASTA specs, but this
behavior can be disabled by specifing the `-w` option.

.. warning::

    The reference file is loaded in memory

dbm command
***********

Creates a dbm DB using the `semidbm` package. The database can then be loaded
using :class:`mgkit.db.dbm.GFFDB`

mongodb command
***************

Outputs annotations in a format supported by MongoDB. More information about it
can be found in :mod:`mgkit.db.mongo`

gtf command
***********

Outputs annotations in the GTF format

split command
*************

Splits a GFF file into smaller chunks, ensuring that all of a sequence
annotations are in the same file.

Changes
*******

.. versionchanged:: 0.3.0
    added *--split* option to *sequence* command

.. versionchanged:: 0.2.6
    added *split* command, *--indent* option to *mongodb*

.. versionchanged:: 0.2.3
    added *--gene-id* option to *gtf* command

.. versionadded:: 0.2.2
    added *gtf* command

.. versionadded:: 0.2.1
    *dbm* and *mongodb* commands

.. versionadded:: 0.1.15

"""

from __future__ import division
import sys
import argparse
import logging
import functools

import mgkit
from . import utils
from mgkit.io import gff, fasta
from mgkit.db import dbm
from mgkit import taxon
from mgkit import simple_cache

LOG = logging.getLogger(__name__)


def set_sequence_parser(parser):
    parser.add_argument(
        '-r',
        '--reverse',
        action='store_true',
        help='Reverse complement sequences on the - strand',
        default=False
    )
    parser.add_argument(
        '-w',
        '--no-wrap',
        action='store_true',
        help='Write the nucleotidic sequence on one line',
        default=False
    )
    parser.add_argument(
        '-s',
        '--split',
        action='store_true',
        help='''Split the sequence header of the reference at the first
        space, to emulate BLAST behaviour''',
        default=False
    )
    parser.add_argument(
        '-f',
        '--reference',
        type=argparse.FileType('r'),
        default=None,
        help='Fasta file containing the reference sequences of the GFF file'
    )

    parser.set_defaults(func=sequence_command)


def sequence_command(options):
    if options.reference is None:
        utils.exit_script('A fasta reference file is required', 1)

    wrap = 60

    if options.no_wrap:
        wrap = None

    seqs = dict(
        (
            seq_id.split(' ')[0] if options.split else seq_id,
            seq
        )
        for seq_id, seq in fasta.load_fasta(options.reference)
    )

    ann_iter = gff.parse_gff(options.input_file, gff_type=gff.from_gff)

    seq_iter = gff.extract_nuc_seqs(ann_iter, seqs, reverse=options.reverse)

    for name, seq in seq_iter:
        fasta.write_fasta_sequence(options.output_file, name, seq, wrap=wrap)


def set_dbm_parser(parser):
    parser.add_argument(
        '-d',
        '--output-dir',
        default='gff-dbm',
        type=str,
        help='Directory for the database'
    )
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input GFF file, defaults to stdin'
    )

    parser.set_defaults(func=dbm_command)


def dbm_command(options):
    db = dbm.create_gff_dbm(
        gff.parse_gff(options.input_file),
        options.output_dir
    )
    db.close()


def set_mongodb_parser(parser):
    parser.add_argument(
        '-t',
        '--taxonomy',
        type=str,
        default=None,
        help='Taxonomy used to populate the lineage'
    )
    parser.add_argument(
        '-c',
        '--no-cache',
        action='store_false',
        default=True,
        help='No cache for the lineage function'
    )
    parser.add_argument(
        '-i',
        '--indent',
        action='store',
        type=int,
        default=None,
        help='No cache for the lineage function'
    )

    parser.set_defaults(func=mongodb_command)


def mongodb_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    lineage_func = None

    if options.taxonomy is not None:
        taxonomy = taxon.UniprotTaxonomy(options.taxonomy)
        lineage_func = functools.partial(
            taxon.get_lineage,
            taxonomy
        )
        if options.no_cache:
            LOG.info('Using cached calls to lineage')
            lineage_func = simple_cache.memoize(lineage_func)

    for annotation in gff.parse_gff(options.input_file):
        options.output_file.write(
            annotation.to_mongodb(
                lineage_func=lineage_func,
                indent=options.indent
            )
        )
        options.output_file.write('\n')


def gtf_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    for annotation in gff.parse_gff(options.input_file):
        options.output_file.write(
            annotation.to_gtf(
                gene_id_attr=options.gene_id
            )
        )


def set_gtf_parser(parser):
    parser.add_argument(
        '-g',
        '--gene-id',
        type=str,
        default='gene_id',
        help='GFF attribute to use for the GTF gene_id attribute'
    )
    parser.set_defaults(func=gtf_command)


def set_common_options(parser):
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input GFF file, defaults to stdin'
    )
    parser.add_argument(
        'output_file',
        nargs='?',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output file, defaults to stdout'
    )


def set_split_parser(parser):
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input GFF file, defaults to stdin'
    )
    parser.add_argument(
        '-p',
        '--prefix',
        type=str,
        default='split',
        help='Prefix for the file name in output'
    )
    parser.add_argument(
        '-n',
        '--number',
        type=int,
        default=10,
        help='Number of chunks into which split the GFF file'
    )
    parser.add_argument(
        '-z',
        '--gzip',
        action='store_true',
        default=False,
        help='gzip output files'
    )
    parser.set_defaults(func=split_command)


def split_command(options):
    LOG.info(
        "Splitting GFF into %d chunks with prefix %s",
        options.number,
        options.prefix
    )

    name_mask = "%s-{0:05}.gff" % options.prefix
    if options.gzip:
        name_mask += '.gz'
        LOG.info("Output files will be compressed (gzip)")

    gff.split_gff_file(
        options.input_file,
        name_mask,
        num_files=options.number
    )


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Extract informations from a GFF file',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()

    parser_s = subparsers.add_parser(
        'sequence',
        help='Extract the nucleotidic sequences of annotations'
    )

    set_sequence_parser(parser_s)
    set_common_options(parser_s)
    utils.add_basic_options(parser_s, manual=__doc__)

    parser_d = subparsers.add_parser(
        'dbm',
        help='Creates a dbm database'
    )

    set_dbm_parser(parser_d)
    utils.add_basic_options(parser_d, manual=__doc__)

    parser_m = subparsers.add_parser(
        'mongodb',
        help='Extract annotations from a GFF file and makes output for MongoDB'
    )

    set_mongodb_parser(parser_m)
    set_common_options(parser_m)
    utils.add_basic_options(parser_m, manual=__doc__)

    parser_gtf = subparsers.add_parser(
        'gtf',
        help='Extract annotations from a GFF file to a GTF file'
    )

    set_gtf_parser(parser_gtf)
    set_common_options(parser_gtf)
    utils.add_basic_options(parser_gtf, manual=__doc__)

    parser_split = subparsers.add_parser(
        'split',
        help='Split annotations from a GFF file to a several files'
    )

    set_split_parser(parser_split)
    utils.add_basic_options(parser_split, manual=__doc__)

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def main():
    "Main function"

    options = set_parser().parse_args()

    mgkit.logger.config_log(options.verbose)
    options.func(options)



# coding=utf8
"""
Simple fasta parser and a few utility functions
"""
import logging
import mgkit.io

LOG = logging.getLogger(__name__)


def load_fasta(f_handle):
    """
    .. versionchanged:: 0.1.13
        now returns uppercase sequences

    Loads a fasta file and returns a generator of tuples in which the first
    element is the name of the sequence and the second the sequence

    Arguments:
        f_handle (str, file): fasta file to open; a file name or a file handle
            is expected

    Yields:
        tuple: first element is the sequence name/header, the second element is
        the sequence
    """
    if isinstance(f_handle, str):
        f_handle = mgkit.io.open_file(f_handle, 'r')

    if getattr(f_handle, 'name', None) is not None:
        LOG.info("Reading fasta file %s", f_handle.name)

    cur_name = ""
    last_name = ""
    nseq = 0
    # Better use ''.join(list) than seq+=seq, it's faster
    cur_seq = []
    # main loop to read file's sequences
    for line in f_handle:
        if line.startswith('>'):
            if cur_seq != []:
                # start of next sequence
                yield cur_name, ''.join(cur_seq).upper()
                nseq += 1
                cur_seq = []
            # save previous name for loop's else clause
            last_name = cur_name
            # classic fasta, all characters besides starting ">" and rightmost
            # whitespace
            cur_name = line[1:].rstrip()
        else:
            cur_seq.append(line.rstrip())
    else:
        # handles cases where there's no newline after last
        # portion of the sequence at EOF
        if nseq != 0:
            if cur_name != last_name:
                yield cur_name, ''.join(cur_seq).upper()
        # case in which only one sequence is present
        else:
            yield cur_name, ''.join(cur_seq).upper()
    f_handle.close()

    LOG.info("Read %d fasta sequences", nseq)


def write_fasta_sequence(file_handle, name, seq, wrap=60, write_mode='a'):
    """
    Write a fasta sequence to file. If the *file_handle* is a string, the file
    will be opened using *write_mode*.

    :param file_handle: file handle or string.
    :param str name: header to write for the sequence
    :param str seq: sequence to write
    :param int wrap: int for the line wrapping. If None, the sequence will be
        written in a single line
    """
    if isinstance(file_handle, str):
        file_handle = open(file_handle, write_mode)

    if wrap is not None:
        seq = '\n'.join(
            seq[pos:pos + wrap]
            for pos in xrange(0, len(seq), wrap)
        )

    file_handle.write(">{0}\n{1}\n".format(name, seq))


def split_fasta_file(file_handle, name_mask, num_files):
    """
    .. versionadded:: 0.1.13

    Splits a fasta file into a series of smaller files.

    Arguments:
        file_handle (file, str): fasta file with the input sequences
        name_mask (str): file name template for the splitted files, more
            informations are found in :func:`mgkit.io.split_write`
        num_files (int): number of files in which to distribute the sequences
    """
    records = load_fasta(file_handle)

    write_func = lambda h, r: write_fasta_sequence(h, *r)

    mgkit.io.split_write(records, name_mask, write_func, num_files=num_files)



"""
Module containing classes and functions to access Kegg data
"""

import logging
import pickle
import re
import itertools
from .utils import dictionary as dict_utils
from .utils.common import deprecated
from .net import uniprot, url_read


LOG = logging.getLogger(__name__)
KEGG_REST_URL = 'http://rest.kegg.jp/'


class KeggCompound(object):
    "Kegg compound"
    # __slots__ = ('cp_id', 'description')

    def __init__(self, cp_id=None, description=''):
        self.cp_id = cp_id
        self.description = description

    def __eq__(self, other):
        """
        >>> KeggCompound('test') == KeggCompound('test')
        True
        >>> KeggCompound('test') == 1
        False
        """
        if not isinstance(other, KeggCompound):
            return False
        return self.cp_id == other.cp_id

    def __ne__(self, other):
        """
        >>> KeggCompound('test') != KeggCompound('test1')
        True
        >>> KeggCompound('test') != 1
        True
        """
        return not self == other

    def __hash__(self):
        return hash(self.cp_id)

    def __str__(self):
        return "{0}: {1}".format(self.cp_id, self.description)

    def __repr__(self):
        return str(self)


class KeggReaction(object):
    "Kegg Reaction"
    # __slots__ = ('rn_id', 'description', 'cp_in', 'cp_out')

    def __init__(self, rn_id=None, description='', cp_in=None, cp_out=None):
        self.rn_id = rn_id
        self.description = description
        self.cp_in = cp_in if cp_in is not None else {}
        self.cp_out = cp_out if cp_out is not None else {}

    def __eq__(self, other):
        """
        >>> KeggReaction('test') == KeggReaction('test')
        True
        >>> KeggReaction('test') == 1
        False
        """
        if not isinstance(other, KeggReaction):
            return False
        return self.rn_id == other.rn_id

    def __ne__(self, other):
        """
        >>> KeggReaction('test') != KeggReaction('test1')
        True
        >>> KeggReaction('test') != 1
        True
        """
        return not self == other

    def __hash__(self):
        return hash(self.rn_id)

    def __str__(self):
        return "{0}: {1} - CPIN ({2}) CPOUT ({3})".format(
            self.rn_id, self.description, len(self.cp_in), len(self.cp_out)
        )

    def __repr__(self):
        return str(self)


class KeggOrtholog(object):
    "Kegg Ortholog gene"
    # __slots__ = ('ko_id', 'description', 'reactions')

    def __init__(self, ko_id=None, description='', reactions=None):
        self.ko_id = ko_id
        self.description = description.replace(';', '')
        self.reactions = reactions if reactions is not None else {}

    def __eq__(self, other):
        """
        >>> KeggOrtholog('test') == KeggOrtholog('test')
        True
        >>> KeggOrtholog('test') == 1
        False
        """
        if not isinstance(other, KeggOrtholog):
            return False
        return self.ko_id == other.ko_id

    def __ne__(self, other):
        """
        >>> KeggOrtholog('test') != KeggOrtholog('test1')
        True
        >>> KeggOrtholog('test') != 1
        True
        """
        return not self == other

    def __getitem__(self, key):
        return self.reactions[key]

    def __setitem__(self, key, value):
        self.reactions[key] = value

    def __hash__(self):
        return hash(self.ko_id)

    def __len__(self):
        return len(self.reactions)

    def __iter__(self):
        for x in self.reactions:
            yield x

    def __str__(self):
        return "{0}: ({1}) {2}".format(self.ko_id, len(self), self.description)

    def __repr__(self):
        return str(self)


class KeggPathway(object):
    "Kegg Pathway"
    # __slots__ = ('path_id', 'description', 'genes')

    def __init__(self, path_id=None, description=None, genes=None):
        self.path_id = path_id
        self.description = description
        self.genes = genes if genes is not None else {}

    def __eq__(self, other):
        """
        >>> KeggPathway('test') == KeggPathway('test')
        True
        >>> KeggPathway('test') == 1
        False
        """
        if not isinstance(other, KeggPathway):
            return False
        return self.path_id == other.path_id

    def __ne__(self, other):
        """
        >>> KeggPathway('test') != KeggPathway('test1')
        True
        >>> KeggPathway('test') != 1
        True
        """
        return not self == other

    def __hash__(self):
        return hash(self.path_id)

    def __contains__(self, item):
        return item in self.genes

    def __getitem__(self, key):
        return self.genes[key]

    def __setitem__(self, key, value):
        self.genes[key] = value

    def __str__(self):
        return "{0} ({2}): {1}".format(
            self.path_id, self.description, len(self)
        )

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.genes)

    def __iter__(self):
        for x in self.genes:
            yield x


class KeggClientRest(object):
    """
    Kegg REST client

    The class includes methods and data to use the REST API provided by Kegg.
    At the moment it provides methods to for 'link', 'list' and 'get'
    operations,

    `Kegg REST API <http://www.kegg.jp/kegg/rest/keggapi.html>`_

    """
    contact = None
    api_url = KEGG_REST_URL
    cpd_re = re.compile(
        r"ENTRY\s+(C\d{5})\s+Compound\nNAME\s+([,.\w+ ()-]+);?"
    )
    rn_name_re = re.compile(r"R\d{5}")
    rn_eq_re = re.compile(r'C\d{5}')
    ko_desc_re = re.compile(
        r"ko:(K\d{5})\t.+?;\s+([\w+, ()/:'\[\]-]+)( \[EC:)?\n?"
    )
    cpd_desc_re = re.compile(
        r"cpd:(C\d{5})\t([\w+, ()\[\]'*.-]+);?\n?"
    )
    id_prefix = {'C': 'cpd', 'k': 'map', 'K': 'ko', 'R': 'rn', 'm': 'path'}

    # Kegg primitives #

    def link_ids(self, target, ids, strip=True, max_len=50):
        """
        The method abstract the use of the 'link' operation in the Kegg API

        The target parameter can be one of the following::

            pathway | brite | module | disease | drug | environ | ko | genome |
            <org> | compound | glycan | reaction | rpair | rclass | enzyme

            <org> = KEGG organism code or T number

        :param str target: the target db
        :param ids: can be either a single id as a string or a list of ids
        :param bool strip: if the prefix (e.g. ko:K00601) should be stripped
        :param int max_len: the maximum number of ids to retrieve with each
            request, should not exceed 50
        :return dict: dictionary mapping requested id to target id(s)
        """
        if isinstance(ids, str):
            ids = [ids]
        if isinstance(ids, (set, frozenset, dict)):
            ids = list(ids)

        data = []
        for idx in range(0, len(ids), max_len):
            if len(ids) > max_len:
                LOG.info(
                    "Downloading links - range %d-%d",
                    idx + 1, idx + max_len
                )
            url = "{0}link/{1}/{2}".format(
                self.api_url, target, '+'.join(ids[idx:idx+max_len])
            )
            t_data = url_read(url, agent=self.contact)
            t_data = t_data.split('\n')
            data.extend(t_data)

        mapping = {}
        for line in data:
            if not line:
                continue
            source, target = line.split()
            if strip:
                source = source.split(':')[1]
                target = target.split(':')[1]
            try:
                mapping[source].append(target)
            except KeyError:
                mapping[source] = [target]

        return mapping

    def list_ids(self, k_id):
        """
        The method abstract the use of the 'list' operation in the Kegg API

        The k_id parameter can be one of the following::

            pathway | brite | module | disease | drug | environ | ko | genome |
            <org> | compound | glycan | reaction | rpair | rclass | enzyme

            <org> = KEGG organism code or T number

        :param str k_id: kegg database to get list of ids
        :return list: list of ids in the specified database
        """
        url = "{0}list/{1}".format(self.api_url, k_id)
        data = url_read(url, agent=self.contact)
        # leave out the last \n
        return data[:-1]

    def get_entry(self, k_id, option=None):
        """
        The method abstract the use of the 'get' operation in the Kegg API

        :param str k_id: kegg id of the resource to get
        :param str option: optional, to specify a format
        """
        url = "{0}get/{1}/{2}".format(
            self.api_url,
            k_id,
            '' if option is None else option
        )
        data = url_read(url, agent=self.contact)
        return data

    def link(self, target, source, options=None):
        """
        .. versionadded:: 0.2.0

        Implements "link" operation in Kegg REST

        http://www.genome.jp/linkdb/
        """
        url = "http://rest.genome.jp/link/{0}/{1}/{2}".format(
            target,
            source,
            '' if options is None else options
        )

        LOG.debug(url)

        data = url_read(url, agent=self.contact)

        mappings = {}
        for line in data.splitlines():
            source_id, target_id, _ = line.rstrip().split('\t')
            source_id = source_id.split(':')[1]
            target_id = target_id.split(':')[1]
            try:
                mappings[source_id].add(target_id)
            except KeyError:
                mappings[source_id] = set([target_id])

        return mappings

    # names #

    def get_ids_names(self, target='ko', strip=True):
        """
        .. versionadded:: 0.1.13

        Returns a dictionary with the names/description of all the id of a
        specific target, (ko, path, cpd, etc.)

        If strip=True the id will stripped of the module abbreviation (e.g.
        md:M00002->M00002)
        """
        id_names = {}

        for line in self.list_ids(target).splitlines():
            kegg_id, name = line.strip().split('\t')
            if strip:
                kegg_id = kegg_id.split(':')[1]
            id_names[kegg_id] = name
        return id_names

    @deprecated
    def get_names(self, k_type, strip=True):
        """
        .. deprecated:: 0.1.13
            use :meth:`KeggClientRest.get_ids_names`

        Returns a dictionary kegg_id->description for the specified kegg module.

        If strip=True the id will stripped of the module abbreviation (e.g.
        md:M00002->M00002)
        """
        return self.get_ids_names(target=k_type, strip=strip)

    @deprecated
    def get_compound_names(self, ids=None):
        """
        .. deprecated:: 0.1.13
            use :meth:`KeggClientRest.get_ids_names`
        """
        if ids is not None:
            return self.get_compound_names(ids)

        return self.get_ids_names('cpd')

    @deprecated
    def get_kos_descriptions(self, rex=False):
        """
        .. deprecated:: 0.1.13
            use :meth:`KeggClientRest.get_ids_names`

        Get all KOs descriptions

        """

        return self.get_ids_names(target='ko')

    def get_ortholog_pathways(self):
        """
        Gets ortholog pathways, replace 'map' with 'ko' in the id
        """
        data = self.get_ids_names('pathway')
        pathways = {}
        for kegg_id, name in data.iteritems():

            kegg_id = kegg_id.replace('map', 'ko')
            pathways[kegg_id] = name

        return pathways

    @deprecated
    def get_compounds_descriptions(self):
        """
        .. deprecated:: 0.1.13
            use *get_compound_names* instead

        Get compound descriptions
        """
        return self.get_ids_names('cpd')

    @deprecated
    def get_reactions_descriptions(self):
        """
        .. deprecated:: 0.1.13
            use *get_compound_names* instead

        Get reaction descriptions
        """
        return self.get_ids_names('reaction')

    # end names #

    def get_reaction_equations(self, ids, max_len=10):
        "Get the equation for the reactions"
        if isinstance(ids, str):
            ids = [ids]
        data = {}
        for idx in range(0, len(ids), max_len):
            if len(ids) > max_len:
                LOG.info(
                    "Downloading reactions equations - range %d-%d",
                    idx + 1, idx + max_len)

            url = "{0}get/{1}".format(
                self.api_url, '+'.join(ids[idx:idx+max_len])
            )

            t_data = url_read(url)
            for entry in t_data.split('///'):
                for line in entry.split('\n'):
                    if line.startswith('EQUATION'):
                        cp_in, cp_out = line.split('<=>')
                        # print cp_in, cp_out
                        cp_in = self.rn_eq_re.findall(cp_in)
                        cp_out = self.rn_eq_re.findall(cp_out)
                        # print cp_in, cp_out
                    elif line.startswith('ENTRY'):
                        # try:
                        name = self.rn_name_re.search(line).group(0)
                        # except:
                        #     print line
                data[name] = {'in': cp_in, 'out': cp_out}

        return data

    def get_pathway_links(self, pathway):
        """
        Returns a dictionary with the mappings KO->compounds for a specific
        Pathway or module
        """
        compounds = self.link_ids('cpd', pathway)
        compounds = set(compounds[pathway])
        kos = self.link_ids('ko', pathway)
        kos = set(kos[pathway])
        cp_rn = self.link_ids('ec', list(compounds))
        ko_rn = self.link_ids('ec', list(kos))
        cp_rn_rev = dict_utils.reverse_mapping(cp_rn)
        ko_cp = dict_utils.combine_dict(ko_rn, cp_rn_rev)

        return ko_cp


class KeggData(object):
    pathways = None
    _ko_map = None
    maps = None

    def __init__(self, fname=None, gen_maps=True):
        self.pathways = {}
        self._ko_map = {}
        self.maps = None
        if fname:
            self.load_data(fname)
            if gen_maps:
                self.gen_ko_map()
                self.gen_maps()

    def get_ko_names(self):
        ko_names = {}
        for path_id in self:
            for ko_id in self[path_id]:
                if ko_id in ko_names:
                    continue
                ko_names[ko_id] = self[path_id][ko_id].description
        return ko_names

    def get_rn_names(self):
        rn_names = {}
        for path_id in self:
            for ko_id in self[path_id]:
                for rn_id in self[path_id][ko_id].reactions:
                    if rn_id in rn_names:
                        continue
                    rn_names[rn_id] = self[path_id][ko_id].reactions[rn_id].description
        return rn_names

    def get_cp_names(self):
        if self.maps is None:
            self.gen_maps()
        return dict(
            (cp_id, cp.description)
            for cp_id, cp in self.maps['cp'].iteritems()
        )

    def get_ko_rn_links(self, path_filter=None, description=False):
        rn_links = {}
        for path_id in self:
            if path_filter is not None:
                if path_id != path_filter:
                    continue
            for ko_id in self[path_id]:
                if description:
                    rn_links[ko_id] = [
                        "{0}: {1}".format(rn.rn_id, rn.description)
                        for rn in self[path_id][ko_id].reactions.itervalues()
                    ]
                else:
                    rn_links[ko_id] = self[path_id][ko_id].reactions.keys()
        return rn_links

    def get_rn_cp_links(self, path_filter=None, description=False):
        rn_links = {}
        for path_id in self:
            if path_filter is not None:
                if path_id != path_filter:
                    continue
            for ko_id in self[path_id]:
                for rn in self[path_id][ko_id].reactions.itervalues():
                    if rn.rn_id in rn_links:
                        continue
                    cp_set = set()
                    if description:
                        cp_set.update(
                            "{0}: {1}".format(cp.cp_id, cp.description)
                            for cp in itertools.chain(
                                rn.cp_in.itervalues(), rn.cp_out.itervalues()
                            )
                        )
                    else:
                        cp_set.update(rn.cp_in, rn.cp_out)
                    rn_links[rn.rn_id] = cp_set
        return rn_links

    def get_ko_cp_links(self, path_filter=None, description=False):
        ko_links = {}
        for path_id in self:
            if path_filter is not None:
                if path_id != path_filter:
                    continue
            for ko_id in self[path_id]:
                if ko_id in ko_links:
                    continue
                cp_set = set()
                for rn in self[path_id][ko_id].reactions.itervalues():
                    if description:
                        cp_set.update(
                            "{0}: {1}".format(
                                cp.cp_id,
                                cp.description
                            )
                            for cp in itertools.chain(
                                rn.cp_in.itervalues(), rn.cp_out.itervalues()
                            )
                        )
                    else:
                        cp_set.update(rn.cp_in, rn.cp_out)
                ko_links[ko_id] = cp_set
        return ko_links

    def get_ko_cp_links_alt(self, direction='out', description=False):
        cp_links = {}
        for path_id in self:
            for ko_id in self[path_id]:
                if ko_id in cp_links:
                    continue
                cp = set()
                for rn in self[path_id][ko_id].reactions.itervalues():
                    cp_dict = getattr(rn, 'cp_' + direction)
                    if description:
                        for cp_obj in cp_dict.itervalues():
                            cp.add(
                                "{0}: {1}".format(
                                    cp_obj.cp_id,
                                    cp_obj.description
                                )
                            )
                    else:
                        cp.update(cp_dict.keys())
                cp_links[ko_id] = cp
        return cp_links

    def gen_ko_map(self):
        for path_id in self:
            for ko_id in self[path_id]:
                try:
                    self._ko_map[ko_id].add(path_id)
                except KeyError:
                    self._ko_map[ko_id] = set()
                    self._ko_map[ko_id].add(path_id)

    def gen_maps(self):
        ko_maps = {}
        rn_maps = {}
        cp_maps = {}
        for path_id in self:
            for ko in self[path_id].genes.itervalues():
                if ko.ko_id in ko_maps:
                    continue
                ko_maps[ko.ko_id] = ko
                for rn in ko.reactions.itervalues():
                    if rn.rn_id in rn_maps:
                        continue
                    rn_maps[rn.rn_id] = rn
                    for cp in itertools.chain(rn.cp_in.values(),
                                              rn.cp_out.values()):
                        if cp.cp_id in cp_maps:
                            continue
                        cp_maps[cp.cp_id] = cp
        self.maps = {'cp': cp_maps, 'ko': ko_maps, 'rn': rn_maps}

    def get_pathway_ko_map(self, black_list=None):
        if black_list is None:
            black_list = self.pathways.keys()
        else:
            # keeps only path_ids that are not in the black_list
            black_list = set(self.pathways.keys()) - set(black_list)

        return dict(
            (path_id, self[path_id].genes.keys())
            for path_id in self
            if path_id in black_list
        )

    def get_ko_pathway_map(self, black_list=None):
        return dict_utils.reverse_mapping(
            self.get_pathway_ko_map(black_list=black_list)
        )

    def get_ko_pathways(self, ko_id):
        return sorted(self._ko_map[ko_id])

    def save_data(self, fname):
        LOG.info("Saving data to file %s", fname)
        pickle.dump(self.pathways, open(fname, 'w'))

    def load_data(self, fname):
        LOG.info("Loading data from file %s", fname)
        self.pathways = pickle.load(open(fname, 'r'))

    def __getitem__(self, key):
        return self.pathways[key]

    def __len__(self):
        return len(self.pathways)

    def __iter__(self):
        for x in self.pathways:
            yield x


class KeggMapperBase(object):
    """
    Base object for Kegg mapping classes
    """
    _ko_map = None
    _not_found = None
    _id_names = None

    def __init__(self, fname=None):
        self._ko_map = {}
        self._not_found = []
        self._id_names = {}
        if fname:
            self.load_data(fname)

    ko_to_mapping = staticmethod(uniprot.ko_to_mapping)

    def save_data(self, fname):
        """
        Saves mapping data to disk
        """
        LOG.info("Saving data to %s", fname)
        pickle.dump((self._ko_map, self._not_found, self._id_names),
                    open(fname, 'w'))

    def load_data(self, fname):
        """
        Loads mapping data to disk
        """
        LOG.info("Loading data from %s", fname)
        data = pickle.load(open(fname, 'r'))
        try:
            self._ko_map, self._not_found, self._id_names = data
        except ValueError:
            LOG.warning("Pickled data in old format")
            self._ko_map, self._not_found = data

    def get_ko_map(self):
        """
        Returns a copy of the KO->mapping dictionary
        """
        return self._ko_map.copy()

    def get_id_map(self):
        """
        Returns a mapping->KOs dictionary (a reverse mapping to get_ko_map)
        """
        return dict_utils.reverse_mapping(self.get_ko_map())

    def get_id_names(self):
        """
        Returns a copy of the mapping names
        """
        return self._id_names.copy()

    def __len__(self):
        return len(self._ko_map)

    def __getitem__(self, key):
        return self._ko_map[key]

    def __setitem__(self, key, value):
        self._ko_map[key] = value

    def __delitem__(self, key):
        del self._ko_map[key]

    def __contains__(self, key):
        return key in self._ko_map

    def __iter__(self):
        return iter(self._ko_map)

BLACK_LIST = [
    'ko05164',  # Influenza A
    'ko05166',  # HTLV-I infection
    'ko05161',  # Hepatitis B
    'ko05160',  # Hepatitis C
    'ko05162',  # Measles
    'ko05169',  # Epstein-Barr virus infection
    'ko05168',  # Herpes simplex infection
    'ko04520',  # Adherens junction
    'ko05014',  # Amyotrophic lateral sclerosis (ALS)
    'ko05016',  # Huntington's disease
    'ko05310',  # Asthma
    'ko00908',  # Zeatin biosynthesis
    'ko04930',  # Type II diabetes mellitus
    'ko04622',  # RIG-I-like receptor signaling pathway
    'ko04626',  # Plant-pathogen interaction
    'ko05222',  # Small cell lung cancer
    'ko05223',  # Non-small cell lung cancer
    'ko05220',  # Chronic myeloid leukemia
    'ko05221',  # Acute myeloid leukemia
    'ko04612',  # Antigen processing and presentation
    'ko04621',  # NOD-like receptor signaling pathway
    'ko04620',  # Toll-like receptor signaling pathway
    'ko04623',  # Cytosolic DNA-sensing pathway
    'ko04622',  # RIG-I-like receptor signaling pathway
    'ko00351',  # DDT degradation
    'ko04270',  # Vascular smooth muscle contraction
    'ko04115',  # p53 signaling pathway
    'ko04114',  # Oocyte meiosis
    'ko05120',  # Epithelial cell signaling in Helicobacter pylori infection
    'ko00590',  # Arachidonic acid metabolism
    'ko04370',  # VEGF signaling pathway
    'ko04976',  # Bile secretion
    'ko04971',  # Gastric acid secretion
    'ko04970',  # Salivary secretion
    'ko04973',  # Carbohydrate digestion and absorption
    'ko04972',  # Pancreatic secretion
    'ko04975',  # Fat digestion and absorption
    'ko04974',  # Protein digestion and absorption
    'ko04977',  # Vitamin digestion and absorption
    'ko04711',  # Circadian rhythm - fly
    'ko04710',  # Circadian rhythm
    'ko04713',  # Circadian entrainment
    'ko04712',  # Circadian rhythm - plant
    'ko04650',  # Natural killer cell mediated cytotoxicity
    'ko04391',  # Hippo signaling pathway - fly
    'ko04151',  # PI3K-Akt signaling pathway
    'ko04150',  # mTOR signaling pathway
    'ko04510',  # Focal adhesion
    'ko04012',  # ErbB signaling pathway
    'ko04013',  # MAPK signaling pathway - fly
    'ko04010',  # MAPK signaling pathway
    'ko03320',  # PPAR signaling pathway
    'ko04530',  # Tight junction
    'ko00981',  # Insect hormone biosynthesis
    'ko00982',  # Drug metabolism - cytochrome P450
    'ko00983',  # Drug metabolism - other enzymes
    'ko05150',  # Staphylococcus aureus infection
    'ko05152',  # Tuberculosis
    'ko04614',  # Renin-angiotensin system
    'ko04610',  # Complement and coagulation cascades
    'ko04340',  # Hedgehog signaling pathway
    'ko00984',  # Steroid degradation
    'ko05010',  # Alzheimer's disease
    'ko05012',  # Parkinson's disease
    'ko04260',  # Cardiac muscle contraction
    'ko04540',  # Gap junction
    'ko05110',  # Vibrio cholerae infection
    'ko05111',  # Vibrio cholerae pathogenic cycle
    'ko04640',  # Hematopoietic cell lineage
    'ko04310',  # Wnt signaling pathway
    'ko04728',  # Dopaminergic synapse
    'ko04724',  # Glutamatergic synapse
    'ko04725',  # Cholinergic synapse
    'ko04726',  # Serotonergic synapse
    'ko04727',  # GABAergic synapse
    'ko04720',  # Long-term potentiation
    'ko04721',  # Synaptic vesicle cycle
    'ko04722',  # Neurotrophin signaling pathway
    'ko04723',  # Retrograde endocannabinoid signaling
    'ko05340',  # Primary immunodeficiency
    'ko05414',  # Dilated cardiomyopathy
    'ko05416',  # Viral myocarditis
    'ko05410',  # Hypertrophic cardiomyopathy (HCM)
    'ko05412',  # Arrhythmogenic right ventricular cardiomyopathy (ARVC)
    'ko04130',  # SNARE interactions in vesicular transport
    'ko05145',  # Toxoplasmosis
    'ko05144',  # Malaria
    'ko04350',  # TGF-beta signaling pathway
    'ko04210',  # Apoptosis
    'ko05200',  # Pathways in cancer
    'ko05202',  # Transcriptional misregulation in cancer
    'ko05203',  # Viral carcinogenesis
    'ko05020',  # Prion diseases
    'ko05143',  # African trypanosomiasis
    'ko05142',  # Chagas disease (American trypanosomiasis)
    'ko05140',  # Leishmaniasis
    'ko00072',  # Synthesis and degradation of ketone bodies
    'ko00073',  # Cutin, suberine and wax biosynthesis
    'ko05330',  # Allograft rejection
    'ko04912',  # GnRH signaling pathway
    'ko05332',  # Graft-versus-host disease
    'ko04910',  # Insulin signaling pathway
    'ko04916',  # Melanogenesis
    'ko04914',  # Progesterone-mediated oocyte maturation
    'ko04672',  # Intestinal immune network for IgA production
    'ko04670',  # Leukocyte transendothelial migration
    'ko04660',  # T cell receptor signaling pathway
    'ko00603',  # Glycosphingolipid biosynthesis - globo series
    'ko00601',  # Glycosphingolipid biosynthesis - lacto and neolacto series
    'ko00600',  # Sphingolipid metabolism
    'ko00604',  # Glycosphingolipid biosynthesis - ganglio series
    'ko04070',  # Phosphatidylinositol signaling system
    'ko04075',  # Plant hormone signal transduction
    'ko05100',  # Bacterial invasion of epithelial cells
    'ko05204',  # Chemical carcinogenesis
    'ko04920',  # Adipocytokine signaling pathway
    'ko04080',  # Neuroactive ligand-receptor interaction
    'ko04320',  # Dorso-ventral axis formation
    'ko04730',  # Long-term depression
    'ko04950',  # Maturity onset diabetes of the young
    'ko04630',  # Jak-STAT signaling pathway
    'ko04742',  # Taste transduction
    'ko04740',  # Olfactory transduction
    'ko04744',  # Phototransduction
    'ko04745',  # Phototransduction - fly
    'ko04512',  # ECM-receptor interaction
    'ko03460',  # Fanconi anemia pathway
    'ko04360',  # Axon guidance
    'ko05219',  # Bladder cancer
    'ko05218',  # Melanoma
    'ko05217',  # Basal cell carcinoma
    'ko05216',  # Thyroid cancer
    'ko05215',  # Prostate cancer
    'ko05214',  # Glioma
    'ko05213',  # Endometrial cancer
    'ko05212',  # Pancreatic cancer
    'ko05211',  # Renal cell carcinoma
    'ko05210',  # Colorectal cancer
    'ko05034',  # Alcoholism
    'ko05033',  # Nicotine addiction
    'ko05032',  # Morphine addiction
    'ko05031',  # Amphetamine addiction
    'ko05030',  # Cocaine addiction
    'ko05134',  # Legionellosis
    'ko05132',  # Salmonella infection
    'ko05133',  # Pertussis
    'ko05130',  # Pathogenic Escherichia coli infection
    'ko05131',  # Shigellosis

    # Too big
    'ko01100',  # Metabolic pathways
    'ko01110',  # Biosynthesis of secondary metabolites
    'ko01120',  # Microbial metabolism in diverse environments

    'ko04662',  # B cell receptor signaling pathway
    'ko04966',  # Collecting duct acid secretion
    'ko05322',  # Systemic lupus erythematosus
    'ko04964',  # Proximal tubule bicarbonate reclamation
    'ko05320',  # Autoimmune thyroid disease
    'ko04962',  # Vasopressin-regulated water reabsorption
    'ko04960',  # Aldosterone-regulated sodium reabsorption
    'ko04961',  # Endocrine and other factor-regulated calcium reabsorption
    'ko04380',  # Osteoclast differentiation
    'ko04664',  # Fc epsilon RI signaling pathway
    'ko04666',  # Fc gamma R-mediated phagocytosis
    'ko04062',  # Chemokine signaling pathway
    'ko04060',  # Cytokine-cytokine receptor interaction
    'ko04066',  # HIF-1 signaling pathway
    'ko04064',  # NF-kappa B signaling pathway
    'ko05323',  # Rheumatoid arthritis
    'ko00140',  # Steroid hormone biosynthesis
]


def download_data(fname='kegg.pickle', contact=None):

    kclient = KeggClientRest()
    kclient.contact = contact
    kdata = KeggData()

    LOG.info("Downloading pathway list")
    path_names = kclient.get_ortholog_pathways()
    LOG.info("Found %d pathways", len(path_names))

    LOG.info("Downloading KO descriptions")
    ko_names = kclient.get_kos_descriptions()
    LOG.info("Found %d KOs", len(ko_names))

    LOG.info("Downloading reactions descriptions")
    rn_names = kclient.get_reactions_descriptions()
    LOG.info("Found %d reactions", len(rn_names))

    LOG.info("Downloading compounds descriptions")
    cpd_names = kclient.get_compounds_descriptions()
    LOG.info("Found %d compounds", len(cpd_names))

    kos = {}
    rns = {}
    cps = {}

    LOG.info("Downloading links pathway-ko (%d)", len(path_names))
    path_links = kclient.link_ids('ko', path_names.keys())
    for path_id, ko_list in path_links.iteritems():
        path = KeggPathway(path_id, path_names[path_id])
        for ko_id in ko_list:
            try:
                name = ko_names[ko_id]
            except KeyError:
                # in this case the actual entry for this gene doesn't exists
                LOG.warning(
                    "KO %s not found in the descriptions, skipping", ko_id
                )
                continue
            try:
                ko = kos[ko_id]
            except KeyError:
                ko = KeggOrtholog(ko_id, name)
                kos[ko_id] = ko
            path[ko_id] = ko
        kdata.pathways[path_id] = path

    LOG.info("Downloading links ko-reactions (%d)", len(ko_names))
    ko_links = kclient.link_ids('rn', ko_names.keys())
    for ko_id, rn_list in ko_links.iteritems():
        try:
            ko = kos[ko_id]
        except KeyError:
            LOG.warning("KO %s is not found in a pathway, skipping", ko_id)
            continue
        for rn_id in rn_list:
            try:
                name = rn_names[rn_id]
            except KeyError:
                LOG.warning(
                    "Reaction %s not found in the descriptions," +
                    " skipping", rn_id)
            try:
                rn = rns[rn_id]
            except KeyError:
                rn = KeggReaction(rn_id, name)
                rns[rn_id] = rn
            ko[rn_id] = rn

    LOG.info("Downloading links reactions-compounds (%d)", len(rn_names))
    cp_links = kclient.get_reaction_equations(rn_names.keys())
    for rn_id, cp_dict in cp_links.iteritems():
        try:
            rn = rns[rn_id]
        except KeyError:
            LOG.warning("Reaction %s is not found in a pathway, skipping",
                        rn_id)
            continue
        for cp_id in cp_dict['in'] + cp_dict['out']:
            try:
                name = cpd_names[cp_id]
            except KeyError:
                LOG.warning("Compound %s not found in the descriptions, " +
                            "skipping", rn_id)
            try:
                cp = cps[cp_id]
            except KeyError:
                cp = KeggCompound(cp_id, name)
                cps[cp_id] = cp
            if (cp_id in cp_dict['in']) and (cp_id in cp_dict['out']):
                LOG.debug(
                    "Compound %s in both side of reaction %s",
                    cp_id, rn_id
                )
            if cp_id in cp_dict['in']:
                rn.cp_in[cp_id] = cp
            if cp_id in cp_dict['out']:
                rn.cp_out[cp_id] = cp

    kdata.save_data(fname)


class KeggModule(object):
    """
    .. versionadded:: 0.1.13

    Used to extract information from a pathway module entry in Kegg

    The entry, as a string, can be either passed at instance creation or with
    :meth:`KeggModule.parse_entry`

    """
    entry = ''
    name = ''
    classes = None
    compounds = None

    _orthologs = None
    _reactions = None
    reactions = None

    def __init__(self, entry=None, old=False):
        """
        .. versionchanged:: 0.3.0
            added *old* parameter, to use the old parser
        """
        if entry is None:
            return
        if old:
            self.parse_entry(entry)
        else:
            self.parse_entry2(entry)

    def parse_entry(self, entry):
        """
        Parses a Kegg module entry and change the instance values. By default
        the reactions IDs are substituted with the KO IDs
        """
        entryd = {}
        curr_field = ''
        for line in entry.splitlines():
            if line.startswith(' '):
                entryd[curr_field].append(line.strip())
            elif line.startswith('///'):
                continue
            else:
                curr_field = line.split(' ')[0]
                entryd[curr_field] = []
                entryd[curr_field].append(line.replace(curr_field, '').strip())

        self.entry = re.search(r"(M\d{5})\s+.+", entryd['ENTRY'][0]).group(1)
        self.name = entryd['NAME'][0]

        self.classes = entryd['CLASS'][0].split('; ')
        self.compounds = [
            re.search(r"(C\d{5})\s+.+", line).group(1)
            for line in entryd['COMPOUND']
        ]

        self.reactions = [
            self.parse_reaction(reaction, ko_ids)
            for ko_ids, reaction in zip(
                entryd['DEFINITION'][0].split(' '),
                entryd['REACTION']
            )
        ]
        self._orthologs = entryd['DEFINITION'][0].split(' ')
        self._reactions = entryd['REACTION']

    def parse_entry2(self, entry):
        """
        .. versionadded:: 0.3.0

        Parses a Kegg module entry and change the instance values. By default
        the reactions IDs are NOT substituted with the KO IDs.
        """
        entryd = {}
        curr_field = ''
        for line in entry.splitlines():
            if line.startswith(' '):
                entryd[curr_field].append(line.strip())
            elif line.startswith('///'):
                continue
            else:
                curr_field = line.split(' ')[0]
                entryd[curr_field] = []
                entryd[curr_field].append(line.replace(curr_field, '').strip())

        self.entry = re.search(r"(M\d{5})\s+.+", entryd['ENTRY'][0]).group(1)
        self.name = entryd['NAME'][0]

        self.classes = entryd['CLASS'][0].split('; ')
        self.compounds = [
            re.search(r"(C\d{5})\s+.+", line).group(1)
            for line in entryd['COMPOUND']
        ]

        self.reactions = []

        for reaction in entryd['REACTION']:
            reaction = self.parse_reaction(reaction, ko_ids=None)
            self.reactions.append(reaction)

        self._orthologs = entryd['DEFINITION'][0].split(' ')
        self._reactions = entryd['REACTION']

    @staticmethod
    def parse_reaction(line, ko_ids=None):
        """
        .. versionchanged:: 0.3.0
            cleaned the parsing

        parses the lines with the reactions and substitute reaction IDs with
        the corresponding KO IDs if provided
        """

        # line = 'R00294,R02492,R09446,R09808,R09809  C00533 -> C00887'
        # ko_ids = '(K00370+K00371+K00374+K00373,K02567+K02568)'
        # some reaction lines have only one space
        rn_ids, reaction = line.replace('  ', ' ').split(' ', 1)
        rn_ids = tuple(x.strip() for x in rn_ids.replace('+', ',').split(','))
        comp1, comp2 = reaction.split(' -> ')
        comp1 = comp1.replace('(spontaneous)', '')
        comp2 = comp2.replace('(spontaneous)', '')
        comp1 = tuple(x.strip() for x in comp1.replace('+', ',').split(','))
        comp2 = tuple(x.strip() for x in comp2.replace('+', ',').split(','))
        if ko_ids is not None:
            rn_ids = ko_ids.replace('+', ',').replace('-', ',').replace('(', '').replace(')', '').split(',')
        return rn_ids, (comp1, comp2)

    @property
    def first_cp(self):
        "Returns the first compound in the module"
        return self.reactions[0][1][0][0]

    @property
    def last_cp(self):
        "Returns the first compound in the module"
        return self.reactions[-1][-1][-1][0]

    def to_edges(self, id_only=None):
        """
        .. versionchanged:: 0.3.0
            added id_only and changed to reflect changes in :attr:`reactions`

        Returns the reactions as edges that can be supplied to make graph.

        Arguments:
            id_only (None, iterable): if None the returned edges are for the
                whole module, if an iterable (converted to a :class:`set`),
                only edges for those reactions are returned

        Yield:
            tuple: the elements are the compounds and reactions in the module
        """

        if id_only is not None:
            id_only = set(id_only)

        for rn_ids, (comp1s, comp2s) in self.reactions:
            for rn_id in rn_ids:
                if (id_only is not None) and (rn_id not in id_only):
                    continue
                for comp1 in comp1s:
                    yield (comp1, rn_id)
                for comp2 in comp2s:
                    yield (rn_id, comp2)

    def find_submodules(self):
        """
        .. versionadded:: 0.3.0

        Returns the possible submodules, as a list of tuples where the elements
        are the first and last compounds in a submodule
        """
        sub_modules = []
        sub_module = None
        for rn_ids, (left_cpds, right_cpds) in self.reactions:
            if sub_module is None:
                sub_module = [left_cpds, right_cpds]
                continue

            if set(sub_module[1]) & set(left_cpds):
                sub_module[1] = right_cpds
            else:
                sub_modules.append((sub_module[0][0], sub_module[-1][-1]))
                sub_module = [left_cpds, right_cpds]
        else:
            sub_modules.append((sub_module[0][0], sub_module[-1][-1]))
        return sub_modules



"""
Utility functions
"""
from __future__ import division
import itertools
import operator
import functools
import warnings


def average_length(a1s, a1e, a2s, a2e):
    """
    Given two sets of coordinates, a1 and a2, returns the average length.

    :param int a1s: a1 leftmost number
    :param int a1e: a1 rightmost number
    :param int a2s: a2 leftmost number
    :param int a2e: a2 rightmost number

    :return float: the average length
    """
    return ((a1e - a1s) + (a2e - a2s) + 2) / 2.0


def between(pos, start, end):
    """
    Tests if a number is between two others

    :param int pos: number to test
    :param int start: leftmost number
    :param int end: rightmost number

    :return bool: if the number is between start and end
    """
    if pos < start or pos > end:
        return False
    return True


def union_range(start1, end1, start2, end2):
    """
    .. versionadded:: 0.1.12

    If two numeric ranges overlap, it returns the new range, otherwise None is
    returned. Works on both int and float numbers, even mixed.

    Arguments:
        start1 (numeric): start of range 1
        end1 (numeric): end of range 1
        start2 (numeric): start of range 2
        end2 (numeric): end of range 2

    Returns:
        (tuple or None): union of the ranges or None if the ranges don't
        overlap

    Example:
        >>> union_range(10, 13, 1, 10)
        (1, 13)
        >>> union_range(1, 10, 11, 13)
        None

    """
    if between(start2, start1, end1) or between(end2, start1, end1):
        return min(start1, start2), max(end1, end2)
    return None


def ranges_length(ranges):
    """
    .. versionadded:: 0.1.12

    Given an iterable where each element is a range, a tuple whose elements
    are numbers with the first being less than or equal to the second, the
    function sums the lengths of all ranges.

    Arguments:
        ranges (iterable): each element is a tuple like *(1, 10)*

    Returns:
        int: sum of all ranges lengths

    """
    return sum(range[1] - range[0] + 1 for range in ranges)


def range_substract(start1, end1, start2, end2):
    intersect = range_intersect(start1, end1, start2, end2)

    if intersect is None:
        return [(start1, end1)]

    ranges = []
    if start1 != intersect[0]:
        ranges.append((start1, intersect[0] - 1))
    if end1 != intersect[1]:
        ranges.append((intersect[1] + 1, end1))
    return ranges


def range_intersect(start1, end1, start2, end2):
    """
    .. versionadded:: 0.1.13

    Given two ranges in the form *(start, end)*, it returns the range
    that is the intersection of the two.

    Arguments:
        start1 (int): start position for the first range
        end1 (int): end position for the first range
        start2 (int): start position for the second range
        end2 (int): end position for the second range

    Returns:
        (None, tuple): returns a tuple with the start and end position for
        the intersection of the two ranges, or *None* if the intersection is
        empty
    """
    if between(start2, start1, end1) or between(end2, start1, end1) or \
       between(start1, start2, end2) or between(end1, start2, end2):
        return max(start1, start2), min(end1, end2)
    return None


def apply_func_window(func, data, window, step=0):

    if step == 0:
        step = window

    for index in xrange(0, len(data), step):
        yield func(data[index:index+window])


def range_substract_(range1, ranges):
    range1 = set(xrange(range1[0], range1[1] + 1))
    range1.difference_update(
        *(
            xrange(x[0], x[1] + 1)
            for x in ranges
        )
    )

    for k, g in itertools.groupby(enumerate(range1), lambda (i, x): i - x):
        group = map(operator.itemgetter(1), g)
        yield min(group), max(group)


def deprecated(func):
    '''
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    from https://wiki.python.org/moin/PythonDecoratorLibrary
    '''

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.\n{}".format(
                func.__name__,
                func.__doc__
            ),
            category=DeprecationWarning,
            filename=func.func_code.co_filename,
            lineno=func.func_code.co_firstlineno + 1
        )
        return func(*args, **kwargs)
    return new_func



"""
Fastq utility functions
"""
import re

CASAVA_HEADER_OLD = r"""(?P<machine>\w+-\w+):
        (?P<lane>\d):
        (?P<tile>\d+):
        (?P<xcoord>\d+):
        (?P<ycoord>\d+)\#
        (?P<index>(\d|[ACTGN]{6}))/
        (?P<mate>(1|2))"""
"Old casava header regex"

CASAVA_HEADER_NEW = r"""(?P<machine>[\w-]+):
        (?P<runid>\d+):
        (?P<cellid>\w+):
        (?P<lane>\d):
        (?P<tile>\d+):
        (?P<xcoord>\d+):
        (?P<ycoord>\d+)
        [_ ](?P<mate>\d): # underscore for data from from www.ebi.ac.uk/ena/
        (?P<filter>[YN]):
        (?P<bits>\d+):
        (?P<index>[ACTGN]+)"""
"New casava header regex"

CASAVA_KHMER = r"""
    (?P<machine>[\w-]+):
    (?P<runid>\d+):
    (?P<cellid>\w+):
    (?P<lane>\d):
    (?P<tile>\d+):
    (?P<xcoord>\d+):
    (?P<ycoord>\d+)
    /(?P<mate>\d)
"""


def check_fastq_type(qualities):
    """
    Trys to guess the type of quality string used in a Fastq file

    :param str qualities: string with the quality scores as in the Fastq file
    :return str: a string with the guessed quality score

    .. note::

        Possible values are the following, classified but the values usually
        used in other softwares:

        * ASCII33: sanger, illumina-1.8
        * ASCII64: illumina-1.3, illumina-1.5, solexa-old

    """
    qualities = [ord(char) for char in qualities]

    max_qual = max(qualities)
    min_qual = min(qualities)

    print min_qual, max_qual

    if (min_qual >= 33) and (max_qual <= 73):
        return 'sanger'
    elif (min_qual >= 33) and (max_qual <= 74):
        return 'illumina-1.8'
    elif (min_qual >= 64) and (max_qual <= 104):
        return 'illumina-1.5'
    elif (min_qual >= 67) and (max_qual <= 104):
        return 'illumina-1.3'
    elif (min_qual >= 59) and (max_qual <= 104):
        return 'solexa-old'


def convert_seqid_to_new(seq_id):
    """
    Convert old seq_id format for Illumina reads to the new found in Casava
    1.8+

    :param str seq_id: seq_id of the sequence (stripped of '@')
    :return str: the new format seq_id

    .. note::

        Example from Wikipedia::

            old casava seq_id:
            @HWUSI-EAS100R:6:73:941:1973#0/1
            new casava seq_id:
            @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCAC

    """
    machine_id, cell_lane, tile_num, x_coord, other = seq_id.split(':')
    y_coord, other = other.split('#')
    idx_multiplex, mate = other.split('/')

    return "{0}:0:FC706VJ:{1}:{2}:{3}:{4} {5}:N:0:{6}".format(
        machine_id,
        cell_lane,
        tile_num,
        x_coord,
        y_coord,
        mate,
        idx_multiplex
    )


def convert_seqid_to_old(seq_id, index_as_seq=True):
    """
    Convert old seq_id format for Illumina reads to the new found in Casava
    until 1.8, which marks the new format.

    :param str seq_id: seq_id of the sequence (stripped of '@')
    :param bool index_as_seq: if True, the index for the multiplex we'll be
        the sequence found at the end of the new format seq_id. Otherwise, 0
        we'll be used
    :return str: the new format seq_id
    """
    left_part, right_part = seq_id.split(' ')
    machine_id, run_id, cell_id, cell_lane, tile_num, x_coord, \
        y_coord = left_part.split(':')

    mate, fail_filter, control_bits, index_seq = right_part.split(':')

    return "HWUSI-{0}:{1}:{2}:{3}:{4}#{5}/{6}".format(
        machine_id,
        cell_lane,
        tile_num,
        x_coord,
        y_coord,
        index_seq if index_as_seq else '0',
        mate
    )


def write_fastq_sequence(file_handle, name, seq, qual, write_mode='a'):
    """
    Write a fastq sequence to file. If the *file_handle* is a string, the file
    will be opened using *write_mode*.

    :param file_handle: file handle or string.
    :param str name: header to write for the sequence
    :param str seq: sequence to write
    :param str qual: quality string
    """
    if isinstance(file_handle, str):
        file_handle = open(file_handle, write_mode)

    file_handle.write(
        "@{name}\n{seq}\n+\n{qual}\n".format(
            name=name,
            seq=seq,
            qual=qual
        )
    )


def choose_header_type(seq_id):
    """
    Return the guessed compiled regular expression
    :param str seq_id: sequence header to test

    :return: compiled regular expression object or None if no match found
    """
    if re.search(CASAVA_HEADER_OLD, seq_id, re.X) is not None:
        return re.compile(CASAVA_HEADER_OLD, re.X)
    if re.search(CASAVA_HEADER_NEW, seq_id, re.X) is not None:
        return re.compile(CASAVA_HEADER_NEW, re.X)
    if re.search(CASAVA_KHMER, seq_id, re.X) is not None:
        return re.compile(CASAVA_KHMER, re.X)
    return None



"""
Module containing classes and functions to deal with CaZy data
"""

import urllib2
import logging
from .. import kegg

CAZY_FAMILIES = {
    'GH': 'Glycoside Hydrolase',
    'GT': 'GlycosylTransferase',
    'PL': 'Polysaccharide Lyase',
    'CE': 'Carbohydrate Esterase',
    'CBM': 'Carbohydrate-Binding Module'
}
"CaZy families"

LOG = logging.getLogger(__name__)


class Kegg2CazyMapper(kegg.KeggMapperBase):
    """
    Class holding mappings KOs->CaZy
    """
    query_string = 'database:(type:ko {0}) AND database:(type:cazy)'
    columns_string = 'database(cazy)'

    def map_kos_cazy(self, kos, contact, skip=False):
        """
        Map a list of KOs to CaZy ids
        """
        msg1 = "(%05d/%05d) Already downloaded mapping for KO: %s - %s"
        msg2 = "(%05d/%05d) Getting Cazy mappings for KO: %s - %s"
        warn1 = "(%05d/%05d) Poblem downloading KO %s (%s)"

        errors = 0
        for idx, ko_id in enumerate(kos):
            if skip and (ko_id in self._ko_map or ko_id in self._not_found):
                LOG.info(msg1, idx + 1, len(kos), ko_id, kos[ko_id])
                continue
            LOG.info(msg2, idx + 1, len(kos), ko_id, kos[ko_id])
            try:
                mappings = self.ko_to_mapping(
                    ko_id, self.query_string,
                    self.columns_string, contact
                )
            except urllib2.HTTPError as url_error:
                LOG.warning(warn1, idx + 1, len(kos), ko_id, url_error.args)
                errors += 1
                continue
            if mappings:
                self._ko_map[ko_id] = mappings
            else:
                self._not_found.append(ko_id)
        return errors


def download_data(contact, kegg_data='kegg.pickle', cazy_data='cazy.pickle'):
    """
    Function to download CaZy data
    """
    WARN1 = "Cazy data not found, download restarting"
    MSG1 = "Loading Kegg data from file %s"
    MSG2 = "Saving downloaded data to %s"
    MSG3 = "Found %d/%d mappings. Number of errors: %d"

    c_obj = Kegg2CazyMapper()
    try:
        c_obj.load_data(cazy_data)
    except IOError:
        LOG.warning(WARN1)

    LOG.info(MSG1, kegg_data)
    kegg_obj = kegg.KeggData(kegg_data)
    ko_names = kegg_obj.get_ko_names()

    try:
        errors = c_obj.map_kos_cazy(ko_names, contact, skip=True)
    finally:
        LOG.info(MSG2, cazy_data)
        c_obj.save_data(cazy_data)

    LOG.info(MSG3, len(c_obj), len(ko_names), errors)



"""
.. versionadded:: 0.1.14

Misc code

"""
import logging

from .. import DependencyError

try:
    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec
except ImportError:
    raise DependencyError('matplotlib')

LOG = logging.getLogger(__name__)


def get_single_figure(dpi=300, figsize=(10, 20), aspect='auto'):
    """
    .. versionchanged:: 0.1.14
        added *aspect* parameter

    Simple wrapper to init a single figure

    Arguments:
        dpi (int): dpi used for the figure
        figsize (tuple): size of the figure in inches
        aspect (str, float): aspect ratio to be passed to figure.add_subplot

    Returns:
        tuple: the figure and axes objects
    """
    fig = plt.figure(dpi=dpi, figsize=figsize)
    ax = fig.add_subplot(111, aspect=aspect)
    return fig, ax


def get_grid_figure(rows, cols, dpi=300, figsize=(10, 20), **kwd):
    """
    .. versionadded:: 0.1.13

    Simple wrapper to init a GridSpec figure

    Arguments:
        rows (int): number of rows
        columns (int): number of columns
        dpi (int): dpi used for the figure
        figsize (tuple): size of the figure in inches

    Returns:
        tuple: the figure and axes objects
    """
    fig = plt.figure(dpi=dpi, figsize=figsize)
    gs = GridSpec(rows, cols, **kwd)
    return fig, gs


__all__ = ['get_grid_figure', 'get_single_figure']



"""
The script contains commands used to access functionality related to
taxonomy, without the need to write ad-hoc code for functionality that
can be part of a workflow. One example is access to the the last common
ancestor function contained in the :mod:`mgkit.taxon`.

lca and lca_line commands
*************************

These commands expose the functionality of
:func:`last_common_ancestor_multiple`, making it accessible via the command
line. They differ in the input file format and the choice of output files.

the *lca* command can be used to define the last common ancestor of contigs
from the annotation in a GFF file. The command uses the *taxon_ids* from all
annotations belonging to a contig/sequence, if they have a **bitscore** higher
or equal to the one passed (50 by default). The default output of the command
is a tab separated file where the first column is the contig/sequence name,
the taxon_id of the last common ancestor, its scientific/common name and its
lineage.

For example::

    contig_21   172788  uncultured phototrophic eukaryote   cellular organisms,environmental samples

If the *-r* is used, by passing the fasta file containing the nucleotide
sequences the output file is a GFF where for each an annotation for the full
contig length contains the same information of the tab separated file format.

The **lca_line** command accept as input a file where each line consist of a
list of taxon_ids. The separator for the list can be changed and it defaults to
TAB. The last common ancestor for all taxa on a line is searched. The ouput of
this command is the same as the tab separated file of the **lca** command, with
the difference that instead of the first column, which in this command becames
a list of all *taxon_ids* that were used to find the last common ancestor for
that line. The list of *taxon_ids* is separated by semicolon ";".

.. note::

    Both also accept the *-n* option, to report the config/line and the
    taxon_ids that had no common ancestors. These are treated as errors and do
    not appear in the output file.

Krona Output
############

**Added in 0.3.0**

The *lca* command supports the writing of a file compatible with Krona. The
output file can be used with the *ktImportText/ImportText.pl* script included
with `KronaTools <https://github.com/marbl/Krona/wiki>`_. Specifically, the
output from *taxon_utils* will be a file with all the lineages found (tab
separated), that can be used with::

    $ ktImportText -q taxon_utils_ouput

Note the use of *-q* to make the script count the lineages. Sequences with no
LCA found will be marked as *No LCA* in the graph, the *-n* is not required.

.. note::

    Please note that the output won't include any sequence that didn't have a
    hit with the software used. If that's important, the **-kt** option can be
    used to add a number of *Unknown* lines at the end, to read the total
    supplied.

Filter by Taxon
***************

The **filter** command of this script allows to filter a GFF file using the
*taxon_id* attribute to include only some annotations, or exclude some. The
filter is based on the `mgkit.taxon.is_ancestor` function, and the
`mgkit.filter.taxon.filter_taxon_by_id_list`. It allows to pass a list of
taxon_id (or taxon_names) to the script. The *include* filter will only output
annotations that have one of the passed taxa as ancestors, while the *exclude*
filter will remove those annotations, that have the passed taxa as ancestors,
from the output.

A list of comma separated taxon_ids can be supplied, as for the names. If any
of the the supplied names have multiple taxon_id (e.g. Actinobacteria) the
script exits and in the log can be found the list of duplicates. For cases like
this, it's preferred for the user to supply a taxon_id, as they can be searched
in NCBI taxonomy (also Uniprot).

.. warning::

    Annotations with no taxon_id are not included in the output of both filters


Changes
*******

.. versionchanged:: 0.3.0
    added *-k* and *-kt* options for Krona output, lineage now includes the LCA
    also added *-a* option to select between lineages with only ranked taxa.
    Now it defaults to all components.

.. versionchanged:: 0.2.6
    added *feat-type* option to *lca* command, added phylum output to nolca

.. versionadded:: 0.2.5

"""
from __future__ import division
import sys
import argparse
import logging
import functools

import mgkit
from . import utils
from mgkit.io import gff, fasta
from mgkit import taxon
from mgkit.simple_cache import memoize
from mgkit.filter.taxon import filter_taxon_by_id_list

LOG = logging.getLogger(__name__)


def set_common_options(parser):
    parser.add_argument(
        '-t',
        '--taxonomy',
        type=str,
        default=None,
        help='Taxonomy file',
        required=True
    )
    parser.add_argument(
        '-k',
        '--krona',
        action='store_true',
        default=False,
        help='Output a file that can be read by Krona (text)',
    )
    parser.add_argument(
        '-kt',
        '--krona-total',
        type=int,
        default=None,
        help='''Total number of raw sequences (used to output correct
                percentages in Krona''',
    )
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input file, defaults to stdin'
    )
    parser.add_argument(
        'output_file',
        nargs='?',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output file, defaults to stdout'
    )


def lca_options(parser):
    parser.add_argument(
        '-n',
        '--no-lca',
        type=argparse.FileType('w'),
        default=None,
        help='File to which write records with no LCA',
    )


def set_lca_contig_parser(parser):
    parser.add_argument(
        '-b',
        '--bitscore',
        default=0,
        type=float,
        help='Minimum bitscore accepted'
    )
    parser.add_argument(
        '-r',
        '--reference',
        default=None,
        type=argparse.FileType('r'),
        help='Reference file for the GFF, if supplied a GFF file is the output'
    )
    parser.add_argument(
        '-s',
        '--sorted',
        action='store_true',
        help='''If the GFF file is sorted (all of a sequence annotations are
                contiguos) can use less memory, `sort -s -k 1,1` can be
                used''',
        default=False
    )
    parser.add_argument(
        '-a',
        '--only-ranked',
        action='store_true',
        help='''If set, only taxa that have a rank will be used in the lineage.
             This is not advised for lineages such as Viruses, where the top
             levels have no rank''',
        default=False
    )
    parser.add_argument(
        '-ft',
        '--feat-type',
        default='LCA',
        type=str,
        help='Feature type used if the output is a GFF (default is *LCA*)'
    )

    lca_options(parser)

    parser.set_defaults(func=lca_contig_command)


def write_lca_gff(file_handle, seq_id, seq, taxon_id, taxon_name, lineage,
                  feat_type):
    annotation = gff.from_sequence(
        seq_id,
        seq,
        feat_type=feat_type,
        taxon_id=taxon_id,
        taxon_name=taxon_name,
        lineage=lineage
    )
    annotation.to_file(file_handle)


def write_lca_tab(file_handle, seq_id, taxon_id, taxon_name, rank, lineage):
    file_handle.write(
        "{}\t{}\t{}\t{}\t{}\n".format(
            seq_id,
            taxon_id,
            taxon_name,
            rank,
            lineage
        )
    )


def get_taxon_info(taxonomy, taxon_id, only_ranked):
    if taxonomy[taxon_id].s_name:
        taxon_name = taxonomy[taxon_id].s_name
    else:
        taxon_name = taxonomy[taxon_id].c_name
    lineage = ','.join(
        tx
        for tx in taxon.get_lineage(taxonomy, taxon_id, names=True, only_ranked=only_ranked, with_last=True)
        if tx
    )
    return taxon_name, lineage


def write_no_lca(file_handle, seq_id, taxon_ids, extra=None):
    file_handle.write(
        "{}\t{}\t{}\n".format(
            seq_id,
            ','.join(str(taxon_id) for taxon_id in set(taxon_ids)),
            '' if extra is None else ','.join(extra)
        )
    )


def write_krona(file_handle, taxonomy, taxon_id, only_ranked):
    if taxon_id is None:
        lineage = ['No LCA']
    else:
        lineage = taxon.get_lineage(
            taxonomy,
            taxon_id,
            names=True,
            only_ranked=only_ranked,
            with_last=True
        )
    file_handle.write(
        '{}\n'.format('\t'.join(lineage))
    )


def lca_contig_command(options):
    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    if options.no_lca is not None:
        LOG.info(
            "Writing contigs/taxon_ids of contigs with no LCA to (%s)",
            getattr(options.no_lca, 'name', repr(options.no_lca))
        )

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    if options.reference is not None:
        seqs = dict(fasta.load_fasta(options.reference))
    else:
        seqs = None

    # basic filter for the presence of a taxon_id and bitscore
    annotations_iter = (
        annotation
        for annotation in gff.parse_gff(options.input_file)
        # only use annotations whose bitscore pass the filter
        # and have a taxon_id
        if (annotation.bitscore >= options.bitscore) and
           (annotation.taxon_id is not None)
    )

    if options.sorted:
        LOG.info("Input GFF is assumed sorted")
        annotations = gff.group_annotations_sorted(
            annotations_iter,
            lambda annotation: annotation.seq_id
        )
    else:
        # groups the annotations by sequence, in case they're not sorted
        annotations = gff.group_annotations(
            annotations_iter,
            lambda annotation: annotation.seq_id
        ).itervalues()

    count = 0
    for seq_ann in annotations:
        count += 1
        seq_id = seq_ann[0].seq_id
        try:
            taxon_id = taxon.last_common_ancestor_multiple(
                taxonomy,
                (annotation.taxon_id for annotation in seq_ann)
            )
        except taxon.NoLcaFound as error:
            LOG.error("No LCA found for %s (%s)", seq_id, error)
            if options.no_lca is not None:
                write_no_lca(
                    options.no_lca,
                    seq_id,
                    (annotation.taxon_id for annotation in seq_ann),
                    extra=set(
                        taxonomy.get_ranked_taxon(
                            annotation.taxon_id,
                            'phylum'
                        ).s_name
                        for annotation in seq_ann
                    )
                )
            if options.krona:
                write_krona(options.output_file, taxonomy, None, False)
            continue

        taxon_name, lineage = get_taxon_info(
            taxonomy,
            taxon_id,
            options.only_ranked
        )
        if seqs is not None:
            write_lca_gff(
                options.output_file,
                seq_id,
                seqs[seq_id],
                taxon_id,
                taxon_name,
                lineage,
                options.feat_type
            )
        elif options.krona:
            write_krona(
                options.output_file,
                taxonomy,
                taxon_id,
                options.only_ranked
            )
        else:
            write_lca_tab(
                options.output_file,
                seq_id,
                taxon_id,
                taxon_name,
                taxonomy[taxon_id].rank,
                lineage
            )
    if options.krona and (options.krona_total is not None):
        for index in xrange(count, options.krona_total):
            options.output_file.write('Unknown\n')


def set_lca_line_parser(parser):
    parser.add_argument(
        '-s',
        '--separator',
        default='\t',
        type=str,
        help='separator for taxon_ids (defaults to TAB)'
    )
    lca_options(parser)
    parser.set_defaults(func=lca_line_command)


def lca_line_command(options):
    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    if options.no_lca is not None:
        LOG.info(
            "Writing taxon_ids of lines with no LCA to (%s)",
            getattr(options.no_lca, 'name', repr(options.no_lca))
        )

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    for line in options.input_file:
        taxon_ids = [
            int(taxon_id)
            for taxon_id in line.strip().split(options.separator)
        ]
        try:
            taxon_id = taxon.last_common_ancestor_multiple(
                taxonomy,
                taxon_ids
            )
        except taxon.NoLcaFound as error:
            LOG.error("No LCA found for %s (%s)", taxon_ids, error)
            if options.no_lca is not None:
                write_no_lca(
                    options.no_lca,
                    '',
                    taxon_ids
                )
            continue
        taxon_name, lineage = get_taxon_info(taxonomy, taxon_id)
        write_lca_tab(
            options.output_file,
            ';'.join(str(x) for x in taxon_ids),
            taxon_id,
            taxon_name,
            taxonomy[taxon_id].rank,
            lineage
        )


def validate_taxon_ids(taxon_ids, taxonomy):
    taxon_ids = set(
        int(taxon_id)
        for taxon_id in taxon_ids.split(',')
        if taxon_id
    )
    broken = False
    for taxon_id in taxon_ids:
        if taxon_id not in taxonomy:
            LOG.critical(
                "taxon_id %d was not found in the taxonomy",
                taxon_id
            )
            broken = True
    if broken:
        utils.exit_script(
            "Some of taxon_ids were not found", 1
        )
    return taxon_ids


def validate_taxon_names(taxon_names, taxonomy):
    taxon_names = tuple(
        taxon_name
        for taxon_name in taxon_names.split(',')
        if taxon_name
    )

    broken = False
    taxon_ids = set()
    for taxon_name in taxon_names:
        try:
            taxon_id = taxonomy.find_by_name(taxon_name)
        except KeyError:
            LOG.critical("'%s' was not found in the taxonomy", taxon_name)
            broken = True
            continue
        if len(taxon_id) > 1:
            LOG.critical(
                "'%s' was found multiple times in the taxonomy: (%s)",
                taxon_name,
                ', '.join(str(tid) for tid in taxon_id)
            )
            broken = True
            continue
        # takes the first (and only) taxon_id from the list returned
        taxon_ids.add(taxon_id[0])

    if broken:
        utils.exit_script(
            "Some of taxon_names were not found", 2
        )
    return taxon_ids


def set_filter_taxa_parser(parser):
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-i',
        '--include-taxon-id',
        action='store',
        type=str,
        help='Include only taxon_ids (comma separated)'
    )
    group.add_argument(
        '-in',
        '--include-taxon-name',
        action='store',
        type=str,
        help='Include only taxon_names (comma separated)'
    )
    group.add_argument(
        '-e',
        '--exclude-taxon-id',
        action='store',
        type=str,
        help='Exclude taxon_ids (comma separated)'
    )
    group.add_argument(
        '-en',
        '--exclude-taxon-name',
        action='store',
        type=str,
        help='Exclude taxon_names (comma separated)'
    )
    parser.set_defaults(func=filter_taxa_command)


def filter_taxa_command(options):
    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    exclude = False

    if options.exclude_taxon_name is not None:
        exclude = True
        taxon_ids = validate_taxon_names(
            options.exclude_taxon_name,
            taxonomy
        )
    elif options.exclude_taxon_id is not None:
        exclude = True
        taxon_ids = validate_taxon_ids(
            options.exclude_taxon_id,
            taxonomy
        )
    elif options.include_taxon_name is not None:
        taxon_ids = validate_taxon_names(
            options.include_taxon_name,
            taxonomy
        )
    elif options.include_taxon_id is not None:
        taxon_ids = validate_taxon_ids(
            options.include_taxon_id,
            taxonomy
        )

    if exclude:
        LOG.info("Excluding Taxa")
    else:
        LOG.info("Only include Taxa")

    comp_taxa = functools.partial(
        filter_taxon_by_id_list,
        filter_list=taxon_ids,
        exclude=exclude,
        func=functools.partial(
            taxon.is_ancestor,
            taxonomy
        )
    )
    comp_taxa = memoize(comp_taxa)

    for annotation in gff.parse_gff(options.input_file):
        if annotation.taxon_id is None:
            continue
        if comp_taxa(annotation.taxon_id):
            annotation.to_file(options.output_file)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Taxonomy Utilities',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()

    parser_lca_contig = subparsers.add_parser(
        'lca',
        help='Finds the last common ancestor for each sequence in a GFF file'
    )

    set_lca_contig_parser(parser_lca_contig)
    set_common_options(parser_lca_contig)
    utils.add_basic_options(parser_lca_contig, manual=__doc__)

    parser_lca_line = subparsers.add_parser(
        'lca_line',
        help='Finds the last common ancestor for all IDs in a text file line'
    )

    set_lca_line_parser(parser_lca_line)
    set_common_options(parser_lca_line)
    utils.add_basic_options(parser_lca_line, manual=__doc__)

    parser_filter_taxa = subparsers.add_parser(
        'filter',
        help='Filter a GFF file based on taxonomy'
    )

    set_filter_taxa_parser(parser_filter_taxa)
    set_common_options(parser_filter_taxa)
    utils.add_basic_options(parser_filter_taxa, manual=__doc__)

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def main():
    "Main function"

    options = set_parser().parse_args()

    mgkit.logger.config_log(options.verbose)
    options.func(options)



"""
Utility functions for the network package
"""

import urllib2
import gzip
import cStringIO
import sys


def url_open(url, data=None, compress=True, agent=None):
    """
    Utility function that compresses the request and add the user agent header
    if supplied.

    :param str url: url to request
    :param str data: data to add to the request
    :param bool compress: if the response should be compressed
    :param str agent: if supplied, the 'User-Agent' header we'll be added to
        the request
    :return: the response handle
    """
    if data is None:
        request = urllib2.Request(url)
    else:
        request = urllib2.Request(url, data)

    if agent is not None:
        request.add_header(
            'User-Agent', "Python {0}.{1} - {2}".format(
                sys.version_info[0],
                sys.version_info[1],
                agent
            )
        )

    if compress:
        request.add_header('Accept-encoding', 'gzip')

    response = urllib2.urlopen(request)

    if response.info().get('Content-Encoding') == 'gzip':
        data_handle = cStringIO.StringIO(response.read())
        data_handle = gzip.GzipFile(fileobj=data_handle, mode='rb')
    else:
        data_handle = response

    return data_handle


def url_read(url, data=None, compress=True, agent=None):
    """
    Utility function that compresses the request and add the user agent header
    if supplied.

    Wrapper of :func:`url_open` which reads the full response

    :param str url: url to request
    :param str data: data to add to the request
    :param bool compress: if the response should be compressed
    :param str agent: if supplied, the 'User-Agent' header we'll be added to
        the request
    :return: the response data
    """
    return url_open(url, data, compress, agent).read()



"""
.. versionadded:: 0.2.6

This script converts annotations in JSON format that were created using MGKit
back into GFF annotations.

mongodb command
***************

Annotations converted into MongoDB records with *get-gff-info mongodb* can be
converted back into a GFF file using this command. It can be useful to get a
GFF file as output from a query to a MongoDB instance on the command line.

"""

from __future__ import division
import argparse
import logging
import json
import sys

import mgkit
from . import utils
from mgkit.io import gff

LOG = logging.getLogger(__name__)


def set_common_options(parser):
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input file, defaults to stdin'
    )
    parser.add_argument(
        'output_file',
        nargs='?',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output file, defaults to stdout'
    )


def set_mongodb_parser(parser):
    parser.set_defaults(func=mongodb_command)


def mongodb_command(options):
    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    LOG.info(
        'Reading from file (%s)',
        getattr(options.input_file, 'name', repr(options.input_file))
    )

    for line in options.input_file:
        annotation = gff.from_mongodb(json.loads(line), lineage=False)
        annotation.to_file(options.output_file)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Convert JSON to GFF',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()

    parser_mongodb = subparsers.add_parser(
        'mongodb',
        help='Convert annotations from a MongoDB instance to GFF'
    )
    set_mongodb_parser(parser_mongodb)
    set_common_options(parser_mongodb)
    utils.add_basic_options(parser_mongodb, manual=__doc__)

    parser_json = subparsers.add_parser(
        'json',
        help='Convert annotations from JSON lines to GFF'
    )
    set_mongodb_parser(parser_json)
    set_common_options(parser_json)
    utils.add_basic_options(parser_json, manual=__doc__)

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def main():
    "Main function"

    options = set_parser().parse_args()

    mgkit.logger.config_log(options.verbose)
    options.func(options)



"""
Adds coverage information to GFF
"""
import argparse
import logging
import pysam
from . import utils
from .. io.gff import parse_gff, write_gff
from .. align import add_coverage_info
from .. import logger
from .add_gff_info import split_sample_alg

LOG = logging.getLogger(__name__)


def set_parser():
    "argument parser configuration"
    parser = argparse.ArgumentParser(
        description='Adds coverage information to GFF files',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        'input_file',
        action='store',
        help='GFF file to annotate',
        type=argparse.FileType('r')
    )
    parser.add_argument(
        'output_file',
        action='store',
        help='output file with coverage information (default: GFF)',
        type=argparse.FileType('w')
    )
    parser.add_argument(
        '-a',
        '--sample-alignment',
        action='append',
        type=split_sample_alg,
        help='sample name and correspondent alignment file separated by comma'
    )
    utils.add_basic_options(parser)

    return parser


def main():
    "Main function"
    options = set_parser().parse_args()

    #configs log and set log level
    logger.config_log(options.verbose)

    samples = []
    bam_files = []

    for sample, bam_file_name in options.sample_alignment:
        samples.append(sample)
        bam_files.append(pysam.Samfile(bam_file_name, 'rb'))

    if len(samples) != len(set(samples)):
        LOG.critical("There are duplicate sample names")
        return 1

    annotations = list(parse_gff(options.input_file))

    add_coverage_info(annotations, bam_files, samples)

    write_gff(annotations, options.output_file)

if __name__ == '__main__':
    main()



"""
Filters GFF annotations in different ways.

Value Filtering
***************

Enables filtering of GFF annotations based on the the first 8 columns, which
are fixed values as well using the last column which holds information in a
key=value way. There are some predefined key=value filters, like `gene_id`,
but `--str-eq`, `--str-in`, `--num-ge` and `--num-le` allow to make additional
filters.

The functions used to make the filters are located in the module
:mod:`mgkit.filter.gff`, and their names start with `filter_base`,
`filter_attr` and `filter_len`.

.. blockdiag::

    {
        orientation = portrait;

        class mgkit [color = "#e41a1c" , textcolor = 'white', width=200,
        fontsize=15];

        class data [color = "#4daf4a" , textcolor = 'white', width=200,
        fontsize=15];
        "GFF" [class = data, shape = flowchart.input];

        parse_gff [class = "mgkit"];
        setup_filters [class = "mgkit"];
        "Filters" [class = "mgkit", stacked, shape = flowchart.condition];

        "GFF" -> parse_gff -> "Filters";
        setup_filters -> Filters;
        "Filtered Annotations" [class = data, stacked];
        "Filters" -> "Filtered Annotations";

    }

Overlap Filtering
*****************

Filters overlapping annotations using the functions
:func:`mgkit.filter.gff.choose_annotation` and
:func:`mgkit.filter.gff.filter_annotations`, after the annotations are grouped
by both sequence and strand. If the GFF is sorted by sequence name and strand,
the `-t` can be used to make the filtering use less memory. It can be sorted in
Unix using `sort -s -k 1,1 -k 7,7 gff_file`, which applies a stable sort using
the sequence name as the first key and the strand as the second key.

.. note::

    It is also recommended to use::

        export LC_ALL=C

    To speed up the sorting

.. blockdiag::

    {
        orientation = portrait;

        class mgkit [color = "#e41a1c" , textcolor = 'white', width=200,
        fontsize=15];
        class data [color = "#4daf4a" , textcolor = 'white', width=200,
        fontsize=15];
        class software [color = "#377eb8", textcolor = "white", fontsize=15];

        sort [class = software];
        "GFF" [class = data, shape = flowchart.input];
        parse_gff [class = "mgkit"];
        group_annotations [class = "mgkit"];
        filter_annotations [class = "mgkit", shape = flowchart.condition,
        fontsize=12];

        "GFF" -> parse_gff -> sort -> filter_annotations;
        parse_gff -> group_annotations -> filter_annotations;

        "Filtered Annotations" [class = data, stacked];
        filter_annotations -> "Filtered Annotations";
    }

The above digram describes the internals of the script.

The annotations needs first to be grouped by seq_id and strand, forming a group
that can be then be passed to :func:`mgkit.filter.gff.filter_annotations`.
This function:

    #. sort annotations by bit score, from the highest to the lowest
    #. loop over all combination of N=2 annotations:

        #. choose which of the two annotations to discard if they overlap for a
           the required amount of bp (defaults to 100bp)
        #. in which case, the preference is given to the db quality first, than
           the bit score and finally the lenght of annotation, the one with the
           highest values is kept

While the default behaviour is the same, now it is posible to decided the
function used to discard one the two annotations. It is possible to use the
`-c` argument to pass a string that defines the function. The string passed must
start with or without a **+**. Using **+** translates into the builtin function
*max* while no **+** translates into *min* from the second character on, any
number of attributes can be used, separated by commas. The attributes, however,
must be one of the properties defined in :class:`mgkit.io.gff.Annotation`,
*bitscore* that returns the value converted in a *float*. Internally the
attributes are stored as strings, so for attributes that have no properties in
the class, such as *evalue*, the `float` builtin is applied.

The tuples built for both annotations are then passed to the comparison
function to be selected and the value returned by it is **discarded**. The
order of the elements in the string is important to define the priority
given to each element in the comparison and the leftmost one has the
highesst priority.

Examples of function strings:

* `-dbq,bitscore,length` becomes max((ann1.dbq, ann1.bitscore, ann1.length),
  (ann2.dbq, ann2.bitscore, ann2.length) - This is default and previously
  only choice
* `-bitscore,length,dbq` uses the same elements but gives lowest priority
  to *dbq*
* `+evalue`: will discard the annotation with the highest *evalue*

Per Sequence Values
*******************

The *sequence* command allows to filter on a per sequence basis, using
functions such as the median, quantile and mean on attributes like evalue,
bitscore and identity. The file can be passed as sorted already, saving memory
(like in the *overlap* command), but it's not needed to sort the file by strand,
only by the first column.

Changes
*******

.. versionadded:: 0.1.12

.. versionchanged:: 0.1.13
    added *--sorted* option

.. versionchanged:: 0.2.0
    changed option *-c* to accept a string to filter overlap

.. versionchanged:: 0.2.5
    added *sequence* command

.. versionchanged:: 0.2.6
    added *length* as attribute and *min*/*max*, and *ge* is the default
    comparison for command *sequence*, *--sort-attr* to *overlap*

"""

import sys
import argparse
import logging
import functools

import pandas

from .. import logger
from . import utils
from ..io import gff
from ..filter import gff as filter_gff

LOG = logging.getLogger(__name__)


def common_options(parser):
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input GFF file, defaults to stdin'
    )
    parser.add_argument(
        'output_file',
        nargs='?',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output GFF file, defaults to stdout'
    )


def parse_attr_arg(value, convert=str):
    values = value.split(':')
    if len(values) != 2:
        raise argparse.ArgumentTypeError(
            "Wrong filter format, must be 'key:value' 'key:value'"
        )
    try:
        values[1] = convert(values[1])
    except ValueError:
        raise argparse.ArgumentTypeError(
            "Wrong filter format, cannot convert {0} to number".format(
                values[1]
            )
        )

    return values[0], values[1]


def set_values_parser(parser):
    parser.add_argument(
        '-s',
        '--seq-id',
        action='store',
        type=str,
        help='filter by sequence id'
    )
    parser.add_argument(
        '--strand',
        action='store',
        type=str,
        help='filter by strand',
        choices=['+', '-']
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-sl',
        '--start-lower',
        action='store',
        type=int,
        help='returns only annotations where the start position is less than'
    )
    group.add_argument(
        '-sg',
        '--start-higher',
        action='store',
        type=int,
        help='''returns only annotations where the start position is greater
                than'''
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-el',
        '--end-lower',
        action='store',
        type=int,
        help='''returns only annotations where the end position is equal or
                less than'''
    )
    group.add_argument(
        '-eg',
        '--end-higher',
        action='store',
        type=int,
        help='''returns only annotations where the end position is equal ot
                greater than'''
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-lg',
        '--length',
        action='store',
        type=int,
        help='filter by annotation length equal to or longer than'
    )
    group.add_argument(
        '-ls',
        '--length-short',
        action='store',
        type=int,
        help='filter by annotation length equal to or shorter than'
    )
    parser.add_argument(
        '--source',
        action='store',
        type=str,
        help='filter by source'
    )
    parser.add_argument(
        '-f',
        '--feat-type',
        action='store',
        type=str,
        help='filter by feature type'
    )
    parser.add_argument(
        '-g',
        '--gene-id',
        action='store',
        type=str,
        help='filter by gene_id'
    )
    parser.add_argument(
        '-d',
        '--db',
        action='store',
        type=str,
        help='filter by db'
    )
    parser.add_argument(
        '-q',
        '--db-qual',
        action='store',
        type=int,
        help='filter by db quality equal or greater than'
    )
    parser.add_argument(
        '-b',
        '--bitscore',
        action='store',
        type=float,
        help='filter by bitscore equal or greater than'
    )
    parser.add_argument(
        '-t',
        '--taxon-id',
        action='store',
        type=int,
        help='filter by taxon_id'
    )
    parser.add_argument(
        '--str-eq',
        action='append',
        type=functools.partial(parse_attr_arg, convert=str),
        help='''filter by custom key:value, if the argument is 'key:value' the
             annotation is kept if it contains an attribute 'key' whose value
             is exactly 'value' as a string.
             '''
    )
    parser.add_argument(
        '--str-in',
        action='append',
        type=functools.partial(parse_attr_arg, convert=str),
        help="Same as '--str-eq' but 'value' is contained in the attribute"
    )
    parser.add_argument(
        '--num-ge',
        action='append',
        type=functools.partial(parse_attr_arg, convert=float),
        help="Same as '--str-eq' but 'value' is a number which is equal or " +
        "greater than"
    )
    parser.add_argument(
        '--num-le',
        action='append',
        type=functools.partial(parse_attr_arg, convert=float),
        help="Same as '--num-ge' but 'value' is a number which is equal " +
        "or less than"
    )

    common_options(parser)

    parser.set_defaults(func=filter_values)


def set_overlap_parser(parser):
    parser.add_argument(
        '-s',
        '--size',
        action='store',
        type=int,
        help='Size of the overlap that triggers the filter',
        default=100
    )
    parser.add_argument(
        '-t',
        '--sorted',
        action='store_true',
        help='''If the GFF file is sorted (all of a sequence annotations are
                contiguos and sorted by strand) can use less memory,
                `sort -s -k 1,1 -k 7,7` can be used''',
        default=False
    )
    parser.add_argument(
        '-c',
        '--choose-func',
        action='store',
        type=make_choose_func,
        help='Function to choose between two overlapping annotations',
        default='dbq,bitscore,length'
    )
    parser.add_argument(
        '-a',
        '--sort-attr',
        action='store',
        help='Attribute to sort annotations before filtering (default ' +
             'bitscore)',
        type=str,
        choices=['bitscore', 'identity', 'length'],
        default='bitscore'
    )

    common_options(parser)

    parser.set_defaults(func=filter_overlaps)


def set_perseq_parser(parser):
    parser.add_argument(
        '-t',
        '--sorted',
        action='store_true',
        help='''If the GFF file is sorted (all of a sequence annotations are
                contiguos) can use less memory, `sort -s -k 1,1` can be used''',
        default=False
    )
    parser.add_argument(
        '-a',
        '--attribute',
        action='store',
        help='Attribute on which to apply the filter',
        type=str,
        choices=['evalue', 'bitscore', 'identity', 'length'],
        default='bitscore'
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-m',
        '--mean',
        action='store_true',
        help='Filter by the mean value',
        default=False
    )
    group.add_argument(
        '-d',
        '--median',
        action='store_true',
        help='Filter by the median value',
        default=False
    )
    group.add_argument(
        '-q',
        '--quantile',
        action='store',
        type=float,
        help='Filter by the quantile value'
    )

    group.add_argument(
        '-s',
        '--std',
        action='store',
        type=float,
        help='''Filter by: mean + (X * std) where X is the number
                supplied'''
    )
    group.add_argument(
        '-x',
        '--max',
        action='store_true',
        help='Filter by the maximum value',
        default=False
    )
    group.add_argument(
        '-n',
        '--min',
        action='store_true',
        help='Filter by the minimum value',
        default=False
    )
    parser.add_argument(
        '-c',
        '--comparison',
        action='store',
        help='Type of comparison (e.g. ge -> greater than or equal to)',
        type=str,
        default='ge',
        choices=['gt', 'ge', 'lt', 'le']
    )

    common_options(parser)

    parser.set_defaults(func=filter_perseq)


def perseq_calc_threshold(annotations, attribute, function, func_arg=None):

    values = pandas.Series(
        annotation.get_attr(attribute, float)
        for annotation in annotations
    )

    if function == 'mean':
        thres = values.mean()
    elif function == 'median':
        thres = values.median()
    elif function == 'quantile':
        thres = values.quantile(func_arg)
    elif function == 'std':
        thres = values.mean() + (func_arg * values.std())
    elif function == 'max':
        thres = values.max()
    elif function == 'min':
        thres = values.min()

    LOG.debug(
        "Threshold for contig %s found: %.2f, using funcion %s and its " +
        "arg %.2f on %d ann.",
        annotations[0],
        thres,
        function,
        func_arg if function in ('quantile', 'std') else 0,
        len(annotations)
    )
    return thres


def find_function(options):
    for function in ('mean', 'quantile', 'median', 'std', 'max', 'min'):
        if getattr(options, function):
            return function
    # in case no function was selected
    return None


def find_comparison(options):
    if options.comparison == 'gt':
        return lambda x, y: x > y
    elif options.comparison == 'ge':
        return lambda x, y: x >= y
    elif options.comparison == 'lt':
        return lambda x, y: x < y
    elif options.comparison == 'le':
        return lambda x, y: x <= y


def filter_perseq(options):

    function = find_function(options)

    if function is None:
        utils.exit_script("No function selected (e.g. mean)", 1)

    comparison = find_comparison(options)

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    file_iterator = gff.parse_gff(options.input_file, gff_type=gff.from_gff)

    if options.sorted:
        LOG.info("Input GFF is assumed sorted")
        grouped = gff.group_annotations_sorted(
            file_iterator, lambda x: x.seq_id
        )
    else:
        grouped = gff.group_annotations(
            file_iterator, lambda x: x.seq_id
        ).itervalues()

    attribute = options.attribute

    for annotations in grouped:
        threshold = perseq_calc_threshold(
            annotations,
            attribute,
            function,
            getattr(options, function)
        )

        for annotation in annotations:
            if comparison(annotation.get_attr(attribute, float), threshold):
                annotation.to_file(options.output_file)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Filter GFF files',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()
    parser_b = subparsers.add_parser(
        'values',
        help='Filter based on values'
    )

    set_values_parser(parser_b)
    utils.add_basic_options(parser_b, manual=__doc__)

    parser_o = subparsers.add_parser(
        'overlap',
        help='Use overlapping filter'
    )

    set_overlap_parser(parser_o)
    utils.add_basic_options(parser_o, manual=__doc__)

    parser_perseq = subparsers.add_parser(
        'sequence',
        help='Filter on a per sequence basis'
    )

    set_perseq_parser(parser_perseq)
    utils.add_basic_options(parser_perseq, manual=__doc__)

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def setup_filters(options):
    filters = []
    # base filters
    base_filters = (
        'taxon_id',
        'seq_id',
        'strand',
        'source',
        'feat_type',
        'gene_id',
        'db'
    )
    for attr in base_filters:
        if getattr(options, attr):
            LOG.info("Filter '%s' = '%s'", attr, getattr(options, attr))
            filters.append(
                functools.partial(
                    filter_gff.filter_base,
                    attr=attr,
                    value=getattr(options, attr)
                )
            )
    # length
    if options.length or options.length_short:

        value = options.length if options.length else options.length_short
        filters.append(
            functools.partial(
                filter_gff.filter_len,
                value=value,
                greater=True if options.length else False
            )
        )
        LOG.info(
            "Filter length %s %s",
            '>=' if options.length else '<=',
            value
        )

    # start position
    if options.start_lower or options.start_higher:
        if options.start_lower:
            value = options.start_lower
        else:
            value = options.start_higher
        filters.append(
            functools.partial(
                filter_gff.filter_base_num,
                attr='start',
                value=value,
                greater=True if options.start_higher else False
            )
        )
        LOG.info(
            "Filter start position %s %s",
            '>=' if options.start_higher else '<=',
            value
        )
    # end position
    if options.end_lower or options.end_higher:
        if options.end_lower:
            value = options.end_lower
        else:
            value = options.end_higher
        filters.append(
            functools.partial(
                filter_gff.filter_base_num,
                attr='end',
                value=value,
                greater=True if options.end_higher else False
            )
        )
        LOG.info(
            "Filter end position %s %s",
            '>=' if options.end_higher else '<=',
            value
        )
    if options.db_qual:
        filters.append(
            functools.partial(
                filter_gff.filter_base_num,
                attr='db_qual',
                value=options.db_qual,
                greater=True
            )
        )
        LOG.info("Filter db quality >= %s", options.db_qual)
    if options.bitscore:
        filters.append(
            functools.partial(
                filter_gff.filter_base_num,
                attr='bitscore',
                value=options.bitscore,
                greater=True
            )
        )
        LOG.info("Filter bitscore >= %s", options.bitscore)
    if options.str_eq:
        for key, value in options.str_eq:
            filters.append(
                functools.partial(
                    filter_gff.filter_attr_str,
                    attr=key,
                    value=value,
                    equal=True
                )
            )
            LOG.info("Filter attribute '%s' = '%s'", key, value)
    if options.str_in:
        for key, value in options.str_in:
            filters.append(
                functools.partial(
                    filter_gff.filter_attr_str,
                    attr=key,
                    value=value,
                    equal=False
                )
            )
            LOG.info("Filter attribute '%s' contains '%s'", key, value)
    if options.num_ge:
        for key, value in options.num_ge:
            filters.append(
                functools.partial(
                    filter_gff.filter_attr_num,
                    attr=key,
                    value=value,
                    greater=True
                )
            )
            LOG.info("Filter attribute '%s' >= %s", key, value)
    if options.num_le:
        for key, value in options.num_le:
            filters.append(
                functools.partial(
                    filter_gff.filter_attr_num,
                    attr=key,
                    value=value,
                    greater=False
                )
            )
            LOG.info("Filter attribute '%s' <= %s", key, value)

    return filters


def filter_values(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    filters = setup_filters(options)

    for annotation in gff.parse_gff(options.input_file, gff_type=gff.from_gff):
        if all(filter(annotation) for filter in filters):
            annotation.to_file(options.output_file)


def make_choose_func(argument):
    """Builds the function used to choose between two annotations."""
    argument = argument.strip()

    LOG.info("Filter function used for overlaps %s", argument)

    if argument.startswith('+'):
        function = max
        argument = argument[1:]
    else:
        function = min

    attributes = argument.split(',')

    choose_func = lambda a1, a2: function(
        a1,
        a2,
        key=lambda el: tuple(
            getattr(el, attribute, None) if hasattr(el, attribute) else el.get_attr(attribute, float)
            for attribute in attributes
        )
    )

    return choose_func


def filter_overlaps(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    file_iterator = gff.parse_gff(options.input_file, gff_type=gff.from_gff)

    if options.sorted:
        LOG.info("Input GFF is assumed sorted")
        grouped = gff.group_annotations_sorted(file_iterator)
    else:
        grouped = gff.group_annotations(file_iterator).itervalues()

    choose_func = functools.partial(
        filter_gff.choose_annotation,
        overlap=options.size,
        choose_func=options.choose_func
    )

    for annotations in grouped:
        filtered = filter_gff.filter_annotations(
            annotations,
            choose_func=choose_func,
            sort_func=lambda x: x.get_attr(options.sort_attr, float),
            reverse=True
        )
        for annotation in filtered:
            annotation.to_file(options.output_file)


def main():
    "Main function"

    options = set_parser().parse_args()

    logger.config_log(options.verbose)

    options.func(options)
